////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MKB;

import java.util.*;
import java.io.*;

/**
 *
 * One Information Source (consists of relations)<BR> This holds info
 * about the IS. DBManagers are no longer accessible from here, rather
 * the Middleware (MW) package is responsible for getting a list of IS
 * objects and generating DBManagers from it. So this is only meta-data
 * @author koeller */

public class IS {

    /** the name of this IS */
    public String ISName=null;
    
    /**
     * this is a vector of objects of type DSRG.MKB.Relation */
    private Vector relations = new Vector(10,10);

  /**
   * Database connection type. For easier initialization of the
   * system.  Now, it could be "ORACLE", and "MS_ACCESS".
   * @author Xin Zhang.  */
  public String DBType;

  /**
   * JDBC URL.  For easier initialization of the system.  */
  public String DBURL;

  /**
   * JDBC Username. For easier initialization of the system.  */
  public String username;

  /**
   * JDBC Password. For easier initialization of the system.
   */
  public String password;

  /**
   * Block Size of that DBMS. Used for QC model.
   * @author Xin Zhang
   */
  int iBF;
    
  /**
   * Constructed used by MKBParser. It used to create an IS object with
   * associated DBManager.
   * @author Xin Zhang.
   */
  public IS (MKB mkb, String ISName,  String sDBType, String sDBURL, 
             String sUsername, String sPassword, int iBF, boolean localIS) {
      this.ISName = ISName;
      DBType = sDBType;
      DBURL = sDBURL;
      username = sUsername;
      password = sPassword;
      this.iBF = iBF;
      if (!localIS) mkb.addIS(this);
      else mkb.setLocalIS(this);
  }
    
    /**
     * returns a vector of the relations in this IS
     * */
    
    public Vector getRelations(){
        return relations;
    }
    

    /**
     * returns a vector of attributes that belong to the relation
     * named in the parameter (in this IS) */

    public Vector getAttributes(String relName) {
        Relation tmpRel = null;
        if ((tmpRel = findRelation(relName)) == null)
            return null;
        return tmpRel.getAttributes();
    }


    /**
     * sets the name of this IS. This replaces the changeName-Method
     in the MKB-package */
    //    public void setName( String newName) {
    //    ISName = newName;
    //}
    
    
    /**
     * returns the name of this IS
     */
    
    public String getName(){
        return ISName;
    }


    /** this method displays a string-representation of this IS. Used
        for testing purposes. It calls the respective
        Relation.displayAll-methods recursively */
    public void displayAll(){
        for(int i=0;i<relations.size();i++){
            System.out.println(ISName+" "+
                               ((Relation)relations.
                                elementAt(i)).displayAll());
        }
    }

  /**
   * Returns a string representation of the entire (MKB) IS object
   * @author xin zhang */
  public String toString(){
    return ISName + relations;
  }
    
    /** returns a relation object that matches the name in the
     * parameter. Replaces getRelation(String).
     * @return null if relation with this name does not exist */
    
    public Relation findRelation(String Name){
        if (relations==null) return null;
        for(int i=0;i<relations.size();i++){
            if(Name.equals(((Relation)relations.elementAt(i)).getName())){
                return (Relation)relations.elementAt(i);
            }
        }
        return null;
    }
    

    /** returns an attribute object after passing relation and
     *  attribute names. 
     *  @return null if nothing found */

    public Attribute findAttribute(String RName,String AttName){
        Relation tmpRel;
        if((tmpRel=findRelation(RName))==null)
            return null;
        return tmpRel.findAttribute(AttName);
        
    }
    
  /**
    * returns a relation object as an answer to the query. <BR>
    * this returns a vector of attributes with vectors of extents attached <BR>
    * they represent a query result from the current DB state in this
    * information source. <B>DISABLED FOR REENGINEERING 111199 koeller</B>
    * @return null is there is no DBManager set
    * */
    
    /*
    public Relation getViewExtent(String SQL) {
        if (dbmgr == null) {
            System.err.println("IS "+ISName+" doesn't have DBManager");
            return null;
        }
        Relation relation = dbmgr.executeViewQuery(SQL);
        return relation;
    }
    */    

  /**
    * returns a RelationExtent object as an answer to the query. <BR>
    * this represents a query result from the current DB state in this
    * information source. This should be called only if this instance
    * is localIS (I think). <B>DISABLED FOR REENGINEERING 111199 koeller</B>
    * @return null is there is no DBManager set
    * */
    /*
      public RelationExtent newGetViewExtent(String SQL) {
      if (dbmgr == null) {
      System.err.println("IS "+ISName+" doesn't have DBManager");
      return null;
      }
      return (dbmgr.newExecuteViewQuery(SQL));
      }
    */
    
    
    /**
     * sends a data update to a certain relation<BR>
     * the format is "value1|value2|...|valuen" <BR>
     * parameters are relationname and count<BR>
     * used for incremental view maintenance<BR><B>
     * DISABLED FOR REENGINEERING 111199 koeller</B>
     *
     */
    /*    
          public boolean sendUpdate(String relName, int Count, String update) {
          return dbmgr.sendUpdate(relName, Count, update);
          }
    */
    
    /** this is for adding empty relation objects during system
     *  initialization. It's called from the MKB parser.  <p> 
     *
     */
    public boolean addRelation (Relation rel) {
        if(findRelation(rel.getName())!=null) return false;
        relations.addElement(rel);
        return true;
    }
    
    /**
     * Add a Relation Object to this IS. Also creates the attribute objects.
     * 
     */
    
    public boolean addRelation ( String relName, Vector attNames, 
                                 Vector attTypes ) {
        
	String tmpName,tmpType;
        
        if (findRelation(relName)!=null) return false;
        
        
        Relation newRel = new Relation(this, relName, -1); // size is unknown
        // relation adds itself to relations-Vector!
        
        // I assume that the two Vectors have the same number of elements
        // create schema
        Enumeration nameEnum = attNames.elements();
        Enumeration typeEnum = attTypes.elements();
        
        
        while (nameEnum.hasMoreElements()) {
            tmpName = (String)nameEnum.nextElement();
            tmpType = (String)typeEnum.nextElement();
            Attribute attr = new Attribute(newRel, tmpName, -1);
            // the attribute adds itself to the relation, so we need no 
            // permanent variable
            attr.setType(tmpType);
        }
        
        return true;
    }

    /**
     * deletes a relation from the relations-Vector. Does nothing
     * else. This should also free the attributes connected to this
     * relation and make them ready for garbage collection. (?) */
    public boolean deleteRelation(Relation R) {
        return relations.removeElement(R);
    }
}
