////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MKB;

import java.util.*;
import java.io.*;

import DSRG.MKB.visitor.MKBVisitor;
import DSRG.MKB.syntaxtree.Node;

//import DSRG.VSA.VSA;
//import DSRG.VKB.View_Definition;
//import DSRG.VSA.Schema_Change;

/** 
 * this is the Meta Knowledge Base main class. It hold lists of ISs,
 * JCs and PCs as well as derived PCs and maintains the MKB under
 * schema changes. */
public class MKB implements Serializable{
  
private static final boolean DEBUG = false;

  /** we need a handle to the VSA to notify it of schema changes,
   *  too. (currently only need after delete-operations. This cannot
   *  be passed in the constructor since it is not known at
   *  construction time. */
    //  private VSA mainVSA;

    private IS localIS; // to park the local-IS handle for later use
 // in the main-class
  
    public Vector IS_List=new Vector();
    public Vector JC_List=new Vector();   // JoinConstraint
    public Vector PC_List=new Vector();   // PCConstraint
    public Vector LC_List=new Vector();   // LocalCondition

    public Vector DerivedPC;
  
  /**
   * A default constructor to create an empty MKB object.
   * It used to let the MKB parser to fill it up!
   * @author Xin Zhang
   */
   public MKB(){
       //    mainVSA = null;
    IS_List = new Vector();
    JC_List = new Vector();
    PC_List = new Vector();
    LC_List = new Vector();
    DerivedPC = null;
  }
  
  
  /**
   * Constructor uses the New MKB Parser.
   * @author Xin Zhang.
   */
  public MKB(String MKBFile){
    try {
      BufferedInputStream Bin = 
	new BufferedInputStream(new FileInputStream (MKBFile));
      MKBParser parser = new MKBParser(Bin);
      parser.disable_tracing();
      Node root = parser.MKBFile();
      System.err.println("MKB parsed successfully.");
      MKBVisitor mkbv = new MKBVisitor();
      root.accept(mkbv);
      // Update the current MKB object 
      update(mkbv.getMKB());
    }
    catch(IOException ioe){
      System.out.println("IOException"+ioe);
    }
    catch(ParseException pe){
        System.out.println("MKB parsed failed.");
        System.out.println("Parse Exception: "+pe.toString());
    }
    
    CMatrix CCC = new CMatrix(this);
    DerivedPC=CCC.GetDerivedPC();
    
    
    // Add derived PCs to list of PCs
    Enumeration pcEnum = DerivedPC.elements();
    PCConstraint tempPC = null;
    while (pcEnum.hasMoreElements()) {
      tempPC = (PCConstraint)pcEnum.nextElement();
      addPCConstraint(tempPC);
      tempPC.getR1().addPCConstraint(tempPC);
      tempPC.getR2().addPCConstraint(tempPC); 
      // this is weird, but that's how he did it...
    }
  }
  
  /**
   * Update MKB Object by another MKB object.
   * @author Xin Zhang.
   */
  private void update(MKB mkb){
      //mainVSA = mkb.mainVSA;
    IS_List = mkb.IS_List;
    JC_List = mkb.JC_List;
    PC_List = mkb.PC_List;
    LC_List = mkb.LC_List;
    DerivedPC = mkb.DerivedPC;    
    localIS = mkb.localIS;
  }

  public void displayAll(){
        System.out.println("********Type constraint********");
        for(int i=0;i<IS_List.size();i++)
            ((IS)IS_List.elementAt(i)).displayAll();

        System.out.println("\n********Join constraint********");
        for(int i=0;i<JC_List.size();i++)
            ((JoinConstraint)JC_List.elementAt(i)).displayAll();


        System.out.println("\n********PC constraint********");
        for(int i=0;i<PC_List.size();i++)
            ((PCConstraint)PC_List.elementAt(i)).displayAll();


        System.out.println("\n********Local Conditions********");
        for(int i=0;i<LC_List.size();i++)
            ((LocalCondition)LC_List.elementAt(i)).toString();

        System.out.println("\n********Derived PC constraint********");
        for(int i=0;i<DerivedPC.size();i++)
            ((PCConstraint)DerivedPC.elementAt(i)).displayAll();
    }


    //    public void setVSA (VSA myVSA) {
    //    mainVSA = myVSA;
    //}


    //========================================================================
    //========================================================================
    //====                          Information Methods                    ===
    //========================================================================
    //========================================================================

    public Vector getAllAttributesAsStrings(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration attrEnum = r.getAttributes().elements();
        Vector returnVec = new Vector();
        Attribute myAttr = null;
        while (attrEnum.hasMoreElements()) {
            myAttr = (Attribute)attrEnum.nextElement();
            returnVec.addElement(myAttr.toString());
        }
        return returnVec;
    }

    public Vector getAllRelationsAsStrings(String isName) {
        IS is = findIS(isName);
        Enumeration attrEnum = is.getRelations().elements();
        Vector returnVec = new Vector();
        Relation myRel = null;
        while (attrEnum.hasMoreElements()) {
            myRel = (Relation)attrEnum.nextElement();
            returnVec.addElement(myRel.toString());
        }
        return returnVec;
    }

    /** returns all TypeConstraint Objects in a String
        representation. Used by user interfaces. */

    public Vector getAllTcsForRelationAsStrings(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration attrEnum = r.getAttributes().elements();
        Vector returnVec = new Vector();
        Attribute myAttr = null;
        while (attrEnum.hasMoreElements()) {
            myAttr = (Attribute)attrEnum.nextElement();
            returnVec.addElement(myAttr.getTypeConstraintAsString());
        }
        return returnVec;
    }

    public Vector getAllTcsForRelation(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration attrEnum = r.getAttributes().elements();
        Vector returnVec = new Vector();
        Attribute myAttr = null;
        while (attrEnum.hasMoreElements()) {
            myAttr = (Attribute)attrEnum.nextElement();
            returnVec.addElement(myAttr.getTypeConstraint());
        }
        return returnVec;
    }


    public Vector getAllJcsForRelationAsStrings(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration joinEnum = r.getJCs().elements();
        Vector returnVec = new Vector();
        JoinConstraint myJoin = null;
        while (joinEnum.hasMoreElements()) {
            myJoin = (JoinConstraint)joinEnum.nextElement();
            returnVec.addElement(myJoin.toString());
        }
        return returnVec;

    }

    public Vector getAllJcsForRelation(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration joinEnum = r.getJCs().elements();
        Vector returnVec = new Vector();
        JoinConstraint myJoin = null;
        while (joinEnum.hasMoreElements()) {
            myJoin = (JoinConstraint)joinEnum.nextElement();
            returnVec.addElement(myJoin);
        }
        return returnVec;

    }

    public Vector getAllPcsForRelationAsStrings(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration PCEnum = r.getPCs().elements();
        Vector returnVec = new Vector();
        PCConstraint myPC = null;
        while (PCEnum.hasMoreElements()) {
            myPC = (PCConstraint)PCEnum.nextElement();
            returnVec.addElement(myPC.toString());
        }
        return returnVec;
    }

    public Vector getAllPcsForRelation(String isName, String relName) {
        Relation r = findRelation(isName, relName);
        Enumeration PCEnum = r.getPCs().elements();
        Vector returnVec = new Vector();
        PCConstraint myPC = null;
        while (PCEnum.hasMoreElements()) {
            myPC = (PCConstraint)PCEnum.nextElement();
            returnVec.addElement(myPC);
        }
        return returnVec;
    }

    public Vector getAllISsAsStrings() {
        Enumeration ISEnum = IS_List.elements();
        Vector returnVec = new Vector();
        IS myIS = null;
        while (ISEnum.hasMoreElements()) {
            myIS = (IS)ISEnum.nextElement();
            returnVec.addElement(myIS.toString());
        }
        return returnVec;
    }
    
    /** 
     * returns a vector is IS-objects (that are already in the MKB)
     */
    public Vector getISs() {
        return IS_List;
    }

  /**
   * Add a new IS object to this MKB.
   * @author Xin Zhang.
   */
  public boolean addIS(IS newIS){
    if(findIS(newIS.getName())!=null) return false;
    IS_List.addElement(newIS);
    return true;
  }

    public void setLocalIS(IS anIS) {
        localIS = anIS;
    }

    public IS getLocalIS() {
        return localIS;
    }

    //=========================================================================
    //=========================================================================
    //====                      Capability Changes                        =====
    //=========================================================================
    //=========================================================================


    /**
     * Create a new Relation in MKB. Used by SimpleNode. not sure why
     * this is here...  */
    /*    public Relation createRelation(String ISName, String relName){
        IS tmpIS;
        if((tmpIS=findIS(ISName))==null) return null;  // IS doesn't exist
        if (!tmpIS.addRelation(relName)) return null;    // Relation can't be created
        return tmpIS.findRelation(relName);
    }
    */

    /** 
     * the actual capability change "addRelation". Only changes the
     * MKB! */

    public boolean addRelation(String ISName, String RName, 
                               Vector attList, Vector typeList) {
        IS the_IS;
        if ((the_IS=findIS(ISName))==null) 
            return false;
        else 
            return the_IS.addRelation(RName, attList, typeList);
    }
    

    /** 
     * the MKB capability change "deleteRelation". Only changes the
     * MKB! */

    public boolean deleteRelation(String ISName,String RName){
        IS the_IS;
        Relation R;

        if((the_IS=findIS(ISName))==null) return false;
        else if((R=the_IS.findRelation(RName))==null) return false;

        return the_IS.deleteRelation(R);
    }

    public boolean changeRelationName(String isName, String relName,
                                      String newName ) {
        
        IS the_IS;
        if((the_IS=findIS(isName))==null) 
            return false;
        else
            return the_IS.findRelation(relName).changeRelationName(newName);
    }

    /** this adds an attribute as a schema change. The size of the
        attribute is hardcoded and set to 37 */

    public boolean addAttribute(String ISName,String RName,
                                String AttName,String AttType){
        IS the_IS;
        Relation R;
        if((the_IS=findIS(ISName))==null) 
            return false;
        else
            if((R=the_IS.findRelation(RName))==null) 
                return false;
            else {
                Attribute attr = new Attribute(R, AttName, -1);
                attr.setType(AttType);
                return true;
            }
    }

    public boolean deleteAttribute(String ISName,String RName,String AttName){
        IS the_IS;
        Relation R;
        Attribute Att;
        if((the_IS=findIS(ISName))==null) return false;
        else if((R=the_IS.findRelation(RName))==null) return false;
        else if((Att=R.findAttribute(AttName))==null) return false;
        deleteAttribute_maintainMKB(Att);
        return R.deleteAttribute(Att);
    }
    
    
    private void deleteAttribute_maintainMKB (Attribute Att) {
        
        for(int i=0;i<PC_List.size();i++){
	    PCConstraint tempPC=(PCConstraint)PC_List.elementAt(i);
            for(int j=0;j<tempPC.getPCPairs().size();j++){
                if(((PCPair)tempPC.getPCPairs().elementAt(j)).containsAtt(Att)){
                    tempPC.getPCPairs().removeElementAt(j);
                    j--;
                }
	    }
	    if(tempPC.Empty()){
                tempPC.deletePCConstraint();
                this.deletePCConstraint(tempPC);
                i--;
	    }
        }
        
        for(int i=0;i<JC_List.size();i++){
	    JoinConstraint tempJC=(JoinConstraint)JC_List.elementAt(i);
	    for(int j=0;j<tempJC.getConditions().size();j++){
                if(((Condition)tempJC.getConditions().elementAt(j)).containsAtt(Att)){
                    tempJC.getConditions().removeElementAt(j);
                    j--;
                }
	    }
	    if(tempJC.Empty()){
                tempJC.deleteJoinConstraint();
                this.deleteJoinConstraint(tempJC);
                i--;
	    }
        }
    }
    
    /**
     * Change a specific Attribute name from "IS.R.A" to "IS.R.B".
     * <p>
     * Note: IS.R.B should not  exist in the MKB already.
     */
    
    public boolean changeAttributeName(String isName, String relName,
                                       String attName, String newName ) {

        Relation tmpRelation = findRelation(isName, relName);
        Attribute tmpAttribute = tmpRelation.findAttribute(attName);
        if (tmpAttribute==null)
            return false;
        else
            return tmpAttribute.changeName(newName);
    }

    public boolean deletePCConstraint(PCConstraint thePC){
        return PC_List.removeElement(thePC);
    }

    public boolean deleteJoinConstraint(JoinConstraint theJC){
        return JC_List.removeElement(theJC);
    }

    public IS findIS(String ISName){
        if (DEBUG) System.err.println("MKB::findIS: Searching for IS "+ISName);
        for(int i=0;i<IS_List.size();i++){
            if(ISName.equals(((IS)IS_List.elementAt(i)).getName())){
                return (IS)IS_List.elementAt(i);
            }
        }
        return null;
    }

    //======================================================================
    //======================================================================
    //==== internal functions =====
    //======================================================================
    //======================================================================

    public Relation findRelation(String ISName,String RName){
        IS the_IS;
        if((the_IS=findIS(ISName))==null)return null;
        return the_IS.findRelation(RName);
    }



     public boolean addJoinConstraint(JoinConstraint theJC){
        JC_List.addElement(theJC);
        return true;
    }

     public boolean addPCConstraint(PCConstraint thePC){
        PC_List.addElement(thePC);
        return true;
    }

     public boolean addLocalCondition(LocalCondition theLC){
        LC_List.addElement(theLC);
        return true;
    }



    /**
     * return a specific attribute.
     */
    public Attribute findAttribute(String ISName,String RName,String AttName){
        IS the_IS;
        if((the_IS=findIS(ISName))==null)return null;
        return the_IS.findAttribute(RName,AttName);
    }


    public JoinConstraint findJC(String ISName1, String RName1, 
                                 String ISName2, String RName2) {
        JoinConstraint tempJC=null, returnJC = null;
        Relation r1, r2;
        for(int i=0;i<JC_List.size();i++) {
            tempJC = (JoinConstraint)JC_List.elementAt(i);
            r1 = tempJC.getR1();
            r2 = tempJC.getR2();
            if (
                (r1.getISName().equals(ISName1) &&
                r1.getName().equals(RName1) &&
                r2.getISName().equals(ISName2) &&
                r2.getName().equals(RName2)) ||
                (r1.getISName().equals(ISName2) &&
                r1.getName().equals(RName2) &&
                r2.getISName().equals(ISName1) &&
                r2.getName().equals(RName1))
                ) {
                returnJC = tempJC;
                break;
            }
        }
        return returnJC;
    }


    /**
     * Find POC PC constraint Pi(upperline{B}(S)) phi Pi(upperline{A}(R))
     * @param rel Relation R
     * @param atleast subset of upperline{A}
     * @param ve View Extent Parameter
     * @return Vector of replacement.
     */

    /*
    public Vector findPOC(Relation rel, Vector atleast, int ve){

        if(DEBUG)
            System.out.println("Debug:MKB:findPOC, rel: " + rel.getFullName() + " atleast: " + atleast);

        Vector result = new Vector();
        DSRG.VSA.Replacement re = null;

        Enumeration e = PC_List.elements();
        PCConstraint pcc;

        // For Each PC Constraint in PC List
        while (e.hasMoreElements()) {
            // create replacement object for this PC constraint
            re = new DSRG.VSA.Replacement();
            pcc = (PCConstraint)e.nextElement();
            if(DEBUG)
                System.out.println("Debug: MKB.findPOC R1:" + pcc.getR1().getName() + " Rel: " + rel.getName());
            // get left relation and try
            if (pcc.getR1()==rel) {
                if(DEBUG)
                    System.out.println("Debug: MKB.findPOC relations matched");
                re.rel = pcc.getR2();
                re.attr_pairs = new Hashtable();
                Enumeration enum = pcc.getPCPairs().elements();
                while(enum.hasMoreElements()){
                    PCPair pp = (PCPair)enum.nextElement();
                    re.attr_pairs.put(pp.getA1(), pp.getA2());
                }
                //match Attributes left
                if(DEBUG)
                    System.out.println("Debug: MKB.findPOC R1 attributes: " + pcc.getR1_Attributes());
                if (VSA.subset(atleast, pcc.getR1_Attributes())){
                    if(DEBUG)
                        System.out.println("Debug: MKB.findPOC R1 attributes matched!");
                    // check ViewExtent parameter
                    switch (ve) {
                    case View_Definition.DONTCARE:
                        result.addElement(re);
                        break;
                    case View_Definition.EQUIVALENT:
                        if(pcc.getExtent().equals("="))
                            result.addElement(re);
                        break;
                    case View_Definition.SUPERSET:
                        if(DEBUG)
                            System.out.println("Debug: MKB.findPOC R1 VE is Superset!");
                        if(DEBUG)
                            System.out.println("Debug: MKB.findPOC PC Op is: " + pcc.getExtent());
                        if(pcc.getExtent().charAt(0) != '>'){
                            result.addElement(re);
                            if(DEBUG)
                                System.out.println("Debug: MKB.findPOC Found PC:" + pcc);
                        }
                        break;
                    case View_Definition.SUBSET:
                        if(pcc.getExtent().charAt(0) != '<')
                            result.addElement(re);
                        break;
                    default:
                        System.out.println("MKB.findPOC ViewExtent Parameter is wrong: " + ve);
                    }
                }
            } // end if R1
            else if (pcc.getR2() == rel) {
                //match Attributes right
                re.rel = pcc.getR1();
                re.attr_pairs = new Hashtable();
                Enumeration enum = pcc.getPCPairs().elements();
                while(enum.hasMoreElements()){
                    PCPair pp = (PCPair)enum.nextElement();
                    re.attr_pairs.put(pp.getA2(), pp.getA1());
                }
                //match Attributes left
                if (VSA.subset(atleast, pcc.getR2_Attributes())){
                    // check ViewExtent parameter
                    switch (ve) {
                    case View_Definition.DONTCARE:
                        result.addElement(re);
                        break;
                    case View_Definition.EQUIVALENT:
                        if(pcc.getExtent().equals("="))
                            result.addElement(re);
                        break;
                    case View_Definition.SUPERSET:
                        if(DEBUG)
                            System.out.println("Debug: MKB.findPOC R2 VE is Superset!");
                        if(pcc.getExtent().charAt(0) != '<')
                            result.addElement(re);
                        break;
                    case View_Definition.SUBSET:
                        if(pcc.getExtent().charAt(0) != '>')
                            result.addElement(re);
                        break;
                    default:
                        System.out.println("MKB.findPOC ViewExtent Parameter is wrong: " + ve);
                    }
                }
            } // end if R2
        }// end While

        if(result.size() == 0)
            System.out.println("MKB.findPOC No PC Found!");
        return result;
    }
    */


  /**
   * Dump this MKB object.
   * @author xin zhang
   */
  public String toString(){
    return IS_List.toString() + "\n"
      + JC_List.toString() + "\n"
      + PC_List.toString();
      
  }


    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================

    //         Additions for QC-Model

    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================
    //=============================================================


    /** 
     * returns relation size (number of tuples) for a given relation
     * Name (dummy values) 
     * @return: -1 if relation Size is unknown */

    public long getRelationSize(String ISName, String relationName) { 
        Relation r = findRelation(ISName, relationName);
 
        if (r!=null) {
            if (DEBUG) System.out.println("QCMODEL: " + 
                        "RelationSize("+ISName+"."+relationName+")= "
                                          + r.getRelationSize());
            return r.getRelationSize();
        }
        else {
            System.out.println("No relation size found for " +
                               ISName + "." + relationName);
            return -1;
        }
    }

    /** returns join selectivity ((R1 \join R2)/(R1 X R2)) for a pair
     *  of given relations (dummy values)
     * @return -1.0 if JS unknown */

    public double getJoinSelectivity(String IS1, String relation1, 
                                  String IS2, String relation2) {

        JoinConstraint jc = findJC(IS1, relation1, IS2, relation2);

        if (jc!=null) {
            if (DEBUG) System.out.println("QCMODEL: " + 
                                          "JS = "+
                                          jc.getJoinSelectivity());

            return jc.getJoinSelectivity();
        }
        else {
            System.out.println("No JC found for "+IS1+"."+
                               relation1+" JOIN "+IS2+"."+relation2);
            return -1.0;
        }
    }


    /**
     * returns local selectivity for a certain local condition. -1 if
     * local condition not found. */
    
    public double getLocalConditionSelectivity(String IS, String relName, 
                                    String attName, String value){
        //dummy function.
        LocalCondition temp;
        if (DEBUG) System.out.println("QCMODEL: Searching for " + 
                                      "LC_S ("+IS
                                      +"."+relName
                                      +"."+attName
                                      +" <=> "+value+")");
        for (int i = 0 ; i < LC_List.size() ; i++ ) {
            // hack: String matching. Couldn't do better right now :)
            temp = (LocalCondition)LC_List.elementAt(i);
            if (temp.getAttribute().indexOf(IS+"."+relName+"."+attName)!=-1)
                if (temp.getValue().indexOf(value)!=-1) {


                    return temp.getSelectivity(); // no values in the grammar :(
                }
        }
        return (double)-1.0;
    }

    /** returns the size of an attribute in bytes. 
     * @return -1 if attribute size unknown */

    public int getAttributeSize(String IS, String relName, String attName) {

        Attribute a = findAttribute(IS, relName, attName);

        if (a!=null) {
            if (DEBUG) System.out.println("QCMODEL: " + 
                        "AttributeSize ("+IS+"."+relName+"."+attName+") = "
                                          + a.getSize());

            return a.getSize();
        }
        else {
            System.out.println("No attribute size found for "
                               + IS + "." + relName + "." + attName);
            return -1;
        }
    }

    /** returns the sum of the the sizes of all attributes in a
     *  relation in bytes. Uses getAttributeSize(String, String,
     *  String)
     * @return -1 if size of at least one attribute is unknown */

    public int getAttributeSizeForRelation(String IS, String relName) {
        Relation r = findRelation(IS,relName);
        Enumeration e = r.getAttributes().elements();
        Attribute a; 
        int size = 0; int attsize;
        while ( e.hasMoreElements()){
            a = (Attribute)e.nextElement();
            attsize = getAttributeSize(IS, relName, a.getName());
            if (attsize == -1) return -1;
            size += attsize;
        }
        if (DEBUG) System.out.println("QCMODEL: "+
                                      "CumulatedAttributeSize ("+IS+"."+relName+") ="+size);
        return size;
    }    
}

