////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MW;

import java.util.*;
import java.sql.*;
import java.io.*;
import DSRG.MKB.*;

public class DBMOracle extends DBManager implements Serializable{

  /**
   *
   *  Oracle specific data types.
   */
  private final String numeric_datatype = "NUMBER";
  private final String date_datatype = "DATE";

  /**
   *  Constructor.
   */
  public DBMOracle(){
    url = "jdbc:oracle:thin:@wpi.wpi.edu:1521:CS";
    className = new String("oracle.jdbc.driver.OracleDriver"); // db driver
    user = new String("eve");
    passwd = new String("eve");
    try{
      driver = Class.forName(className);
      connection = java.sql.DriverManager.getConnection(url, user, passwd);
      System.out.println("Connected");
      System.out.println();
    }
    catch (Exception exc) {
      System.out.println("DB Exception Oracle 1");
      return;
    }
    connected = true;
  }

  /**
   *  Constructor that takes input for the user and passwd... rather than
   *  using the dflt eve, eve;
   */
  public DBMOracle(String myISName, String usr, String pswd, String URL){
      ISName = myISName;
      url = URL; //"jdbc:oracle:thin:@wpi.wpi.edu:1521:CS";
      user = new String(usr);
      passwd = new String(pswd);
      className = new String("oracle.jdbc.driver.OracleDriver"); // db driver
      try{
          driver = Class.forName(className);
          connection = java.sql.DriverManager.getConnection(url, user, passwd);
          System.out.println("Connected");
          System.out.println();
      }
    catch (Exception exc) {
        System.out.println("DB Exception Oracle #2: "+exc.toString());      
        return;
    }
      connected = true;
  }
    
    
    /**
  /**
   * addRelation. <BR> creates only the schema, does not insert data
   * or anything <BR> The types have to be like "VARCHAR2(.."
   * otherwise they're overwritten by "VARCHAR2(31)
   *
   * 10/9/97 koeller */
    public boolean addRelation (String relName,
                                Vector attNames,
                                Vector attTypes) {
        
        java.sql.PreparedStatement stmt = null;
        boolean result = false;
        String schema=new String("(");
        
	// I assume that the two Vectors have the same number of elements
        // create schema
        Enumeration nameEnum = attNames.elements();
        Enumeration typeEnum = attTypes.elements();
        String tmpName, tmpType;
        
        
        while (nameEnum.hasMoreElements()) {
            tmpName = (String)nameEnum.nextElement();
            tmpType = (String)typeEnum.nextElement();
          // try to find the "varchar2("-substring
            if (tmpType.toUpperCase().indexOf("VARCHAR2(")==-1)
                tmpType = "VARCHAR2(31)";
	    schema += (tmpName
                       + " "
                       + tmpType
                       + ", ");
	}
	
	schema = schema.substring(0,schema.length()-2)+")"; 
	// replace the comma at the end with a )
        
        
        // no preparedStatement possible, since Oracle doesn't understand
        // identifier names as parameters...
        // But: Access is the same
        
        try {
            
            stmt=connection.prepareStatement( "create table "
                                              + relName
                                              + " "
                                              + schema);

            if (DEBUG) System.out.println("DBMOracle.addRelation(): "+schema);
            
            stmt.executeUpdate();
            stmt.close ();
            result = true;

        }
        catch (java.sql.SQLException exc) {
            System.out.println("SQL Exception in addRelation: " 
                               + exc.toString());
        }
        catch (Exception e) {
            e.printStackTrace() ;
        }
        
        //    if (result && DB.equals("Access"))
        //    addRelationAcc(RelationName, schema);
        // maintain Meta Tables for Access
        
        
        return result;
    }
    


  /**
   * deleteRelation. <BR>
   *
   * @author 10/6/97 koeller
   * @param name the name of the relation to be deleted
   * @param silent a flag that shows if the error message for deleting
   * a non-existent relation should be suppressed.  */


  public boolean deleteRelation (String name, boolean silent) {

    java.sql.PreparedStatement stmt = null;

    boolean result = false;

    // no preparedStatement possible, since Oracle doesn't understand identifier names
    // as parameters...
    // same thing with Access

    try {


      stmt =
          connection.prepareStatement ( "drop table "
                                        + name);
      
      stmt.executeUpdate ();
      
      
      stmt.close ();
      result = true;
    }
    catch (java.sql.SQLException exc) {
       if (!silent) System.out.println("SQL Exception in deleteRelation: " + exc.toString());
    }


//    if (result && DB.equals("Access")) deleteRelationAcc(name);
    // maintain Meta Tables for Access

    return result;
  }





  /**
   * renameAttribute. <BR>
   *
   *
   * 10/6/97 koeller
   */


  public boolean renameAttribute (String RelationName, String oldName,
				  String newName) {

    java.sql.PreparedStatement stmt = null;

    boolean result = true;

    // no preparedStatement possible, since Oracle doesn't understand
    // identifier names as parameters...

    /* Algo for Oracle. We'll do the same with Access, to get the column
       order right.

       1. get Column names
       2. replace Attribute name; for Access, these are two steps (it doesn't
          do "create as")
       3. create new table with replaced attribute name
       4. drop old table
       5. recreate old table from new table
       6. drop new table

       5.+6. is a renameRelation
     */


    String origString   = getSchema(RelationName);
    String selectString = makeSelectClause(origString);
    String createString = makeSelectClause(renameAttributeInSchema(origString, oldName, newName));


    String makeTable =   "create table "
                 + "temp ("
                 + createString
                 + ") as select "
                 + selectString
                 + " from "
                 + RelationName;


    // 1. drop temp table, just in case

    try {

      stmt=connection.prepareStatement("drop table temp");
      stmt.executeUpdate();
      stmt.close();
    }
    catch (java.sql.SQLException exc) {
    } // ignore SQL errors



    try {
      // 2. make new table "temp"
      stmt=connection.prepareStatement(makeTable);
      stmt.executeUpdate();
      stmt.close();

      // 3. drop old table
      stmt=connection.prepareStatement("drop table "+RelationName);
      stmt.executeUpdate();


      // 4. rename temp table to old name

      if (!renameRelation("temp", RelationName))
	result = false;
      stmt.close();
    }
    catch (java.sql.SQLException exc) {
      result = false;
      System.out.println("SQL Exception in renameAttribute: " + exc.toString());
    }


    return result;
  }



  /**
   * addAttribute. <BR>
   * Extents are being ignored, not needed/possible for the demo <BR>
   *
   *
   * 10/6/97 koeller
   */


  public boolean addAttribute (String RelationName, String AttributeName,
                               String type, int size) {
    java.sql.PreparedStatement stmt = null;

    boolean result = false;

    // no preparedStatement possible, since Oracle doesn't understand
    // identifier names as parameters...

    try {

      stmt=connection.prepareStatement(   "alter table "
                                        + RelationName
                                        + " add "
                                        + AttributeName
                                        + " "
                                        + type
                                        + "("
                                        + (new Integer(size)).toString()
                                        + ")");

      if (DEBUG) System.out.println("DBMOracle.addAttribute :"+RelationName+" "+AttributeName+" "+type+" "+(new Integer(size)));

      stmt.executeUpdate();

      stmt.close ();
      result = true;

    }
    catch (java.sql.SQLException exc) {
       System.out.println("SQL Exception in addAttribute: " + exc.toString());
    }
    return result;
  }




  /**
   * deleteAttribute. <BR>
   * This overwrites any table in the database that is called "tmp31415926".
   *
   * 10/6/97 koeller
   */


  public boolean deleteAttribute (String RelationName, String attributeName) {

     // the only thing that is easier in Access, compared to Oracle
    java.sql.PreparedStatement stmt = null;

    boolean result = true;

      String createString = stripAttributeFromSchema(getSchema(RelationName),
						     attributeName);

      // if we are trying to delete the last attribute ==> error
      if (createString.equals(")")) {
          System.err.println("DBMOracle.deleteAttribute: Trying to delete last attribute.");
          return false;
      }
      String selectString = makeSelectClause(createString);
      String makeTable =   "create table "
                         + "tmp31415926"
                         + " as select "
                         + selectString
                         + " from "
                         + RelationName;


      // 1. drop temp table, just in case

      try {

        stmt=connection.prepareStatement("drop table tmp31415926");
        stmt.executeUpdate();
	stmt.close();
      }
      catch (java.sql.SQLException exc) {
      } // ignore SQL errors

      try {
        // 2. make new table "tmp31415926"

        stmt=connection.prepareStatement(makeTable);
        stmt.executeUpdate();
	stmt.close();

        // 3. drop old table

        stmt=connection.prepareStatement("drop table "+RelationName);
        stmt.executeUpdate();

        // 4. rename temp table to old name

        if (!renameRelation("tmp31415926", RelationName))
             result = false;

	stmt.close();
      }
      catch (java.sql.SQLException exc) {
        result = false;
        System.out.println("SQL Exception in deleteAttribute: " + exc.toString());

      }

      return result;
  }








/**
  *
  * renameRelation <BR>
  *
  * 2/4/98 koeller
  *
  */

  public boolean renameRelation (String oldName, String newName) {
    java.sql.PreparedStatement stmt = null;
    boolean result = false;

    try {
      stmt =
	connection.prepareStatement (   "create table "
					+  newName
					+ " as "
					+ "select * from "
					+ oldName);

      stmt.executeUpdate ();

      stmt =
	connection.prepareStatement (   "drop table "
					+ oldName);

      stmt.executeUpdate ();


      result = true;
      stmt.close();
    }
    catch (java.sql.SQLException exc) {
       System.out.println("SQL Exception in renameRelation: "
			  + exc.toString());
    }
    return result;
  }

/**
  *
  * stores a Table object in a database. <BR>
  * Creates a schema and stores data in there.
  * Uses the relation's name and attribute names and lengths
  * everything is created as "VARCHAR2" !!
  * This is for view recomputation and should only be used on LocalIS
  *
  * @author 04/07/1999 koeller
  *
  */


    public boolean storeRelation(Table table) {

        //System.out.println("Storing Relation: "+ relation.getName());

        Integer aSize = null;
        String schema = new String();
        String tuple = new String();
        Vector column = new Vector();
        boolean result;
        java.sql.PreparedStatement stmt = null;
        String relationName = table.getRelName();

        if (DEBUG) System.out.println("DBMOracle.newStoreRelation: Trying to delete Relation"+relationName);
        deleteRelation(relationName,true);  
        // silent delete

        Enumeration attributeSizes = 
            table.getAttributeSizes().elements();
        Vector attTypes = new Vector();
        while (attributeSizes.hasMoreElements()) {
            aSize = (Integer)attributeSizes.nextElement();
            if (aSize.intValue()==0) aSize = new Integer(29); 
            // preventive maintenance...
            attTypes.addElement("VARCHAR2("+aSize.toString()+")");
        }
        result = addRelation(relationName, table.getAttributeNames(),
                             attTypes); 
     
        if (result) {
            // insert tuples
            Vector rev = table.getTuples();
            if (rev == null) return true;
            if (rev.size()==0) return true;

            try {
                if (rev != null) {
                    Enumeration re = rev.elements();
                    while (re.hasMoreElements()) {
                        tuple = ((Tuple)re.nextElement()).toSQLString();
                    
                        stmt =
                            connection.prepareStatement (   "insert into "
                                                            +  relationName
                                                            + " values "
                                                            + "("
                                                            + tuple
                                                            + ")" );
                        if (DEBUG) System.out.println("DEBUG: inserting: "+   relationName
                                                      + " values "
                                                      + "("
                                                      + tuple
                                                      + ")");
                    
                        stmt.executeUpdate ();
                        stmt.close ();
                    }
                }
	   
            }
            catch (java.sql.SQLException exc) {
                System.out.println("SQL Exception in storeRelation: "
                                   + exc.toString());
                result = false;
            }
        }
        return result;
    }

//====================================================================================
//================                private functions                  =================
//====================================================================================
//====================================================================================
//====================================================================================
//====================================================================================
//====================================================================================

 /**
   *  Helper function. <BR>
   *  This will read the Schema of a given table in the IS, concatenating it to
   *  a string of the form
   *  <code>([</lt>columnname</gt> </lt>type</gt>,] </lt>columnname</gt> </lt>type</gt>)</code>
   *
   */

   private String getSchema(String relationName) {

    java.sql.PreparedStatement stmt = null;
    java.sql.ResultSet rs = null;
    java.sql.ResultSetMetaData rsmt = null;


    String schema = new String("("); // open parenthese
    String result=null;

    String column_name = null;
    String data_type = null;
    String data_length = null;

     try {
      stmt =
	connection.prepareStatement (   "select column_name, data_type, data_length "
                                      + "from cols "
                                      + "where table_name =?");

      stmt.setString(1, relationName);

     // right now, this takes no fixed-precision numbers (like NUMBER(10,2))!!!
     // it only supports simple tables with no indices, foreign-key-constraints,
     // keys etc. just simple datatypes and lengths



      rs = stmt.executeQuery ();
      rsmt = rs.getMetaData ();

      while (rs.next ()) {

        column_name = rs.getString(1).trim();
        data_type = rs.getString(2).trim();
        data_length = rs.getString(3).trim();


        schema= schema + column_name + " " + data_type;

          // Another patch for Access, this is a big mess!!!  //???? koeller 6/2
          //          this is DBMOracle ?!?!?!?!?

        if (!data_type.equals(numeric_datatype))
           schema = schema + "("
                           + data_length
                           + ")";

        schema = schema + ",";
      }

      result= new String(schema.substring(0,schema.length()-1)+")");
      // replace last comma with a ")"

      rs.close ();
      stmt.close ();
    }
    catch (java.sql.SQLException exc) {
       System.out.println("SQL Exception in getAttributeExtent: " + exc.toString());
    }

    return result;
   }

  /**
   * 
   * Reset the Oracle database. this is a hack for the EVE-Demo. It
   * runs a SQL-script whose name is defined in the eve.properties. */
    public void resetDB(String filename){
        try {
            // Open the reset.sql file
            File f  = new File(filename);
            BufferedReader in = new BufferedReader( new FileReader(f));
            Statement stmt = connection.createStatement();
            String sql = null;
            do{
                try{
                    sql = in.readLine();
                    if (DEBUG) System.out.println("DBMOracle.resetDB(). Trying to execute: " + sql);
                    if(sql == null)
	    break;
                    stmt.execute(sql);
                    if (DEBUG) System.out.println("DBMOracle.resetDB() " + sql + " executed!");
                }catch(Exception e){
                    if (DEBUG) System.out.println("DBMOracle.resetDB() inner loop: " + e.getMessage());
                }
            }while(sql!=null);
            stmt.close();
        }catch(Exception e){
      if (DEBUG) System.out.println("DBMOracle.resetDB() outer loop: " + e.getMessage());
        }
    }
}
