////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MW;

import java.util.Vector;
import java.util.Enumeration;
import DSRG.MKB.*;
//import EVEDemo.MW.*;

/** this class holds a number of tuples together with a Relation
 * object in the MKB to form a relation extent. It's used for holding
 * relation extents, since those belong in the MKB.
 * @author koeller (04/07/1999) */

public class Table {


      
    /**
     * Debug flag
     */
    public static final boolean DEBUG = false;
    

    /**
     * the pointer to my Relation in the MKB. Through this relation
     * object, we can get to the Attribute types which should
     * correspond with the types in the tuples.  */
    //    private Relation peerRelation;


    /** a number of tuples in this extent */
    private Vector tuples;

    /** the schema of the Table. A vector of Strings
        (attribute names) */
    private Vector attributeNames;

    /** the sizes of attribute in the attributes of the
        Table. A vector of Integer (lengths) */
    private Vector attributeSizes;


    /** the virtual name of this relation (for query breakdown) */
    private String relName;

    /** construction an empty relation extent with a pointer to an MKB
     *   relation object */
    /*
    public Table(Relation myPeer) {
        peerRelation = myPeer;
        }*/


    /** a new constructor without argument... trying to get rid of the
        peer object */

    public Table() {
        tuples  = new Vector();
    }

    /** constructs a relation extent object with a default size. This
     *   is good for efficiency (if the size of the relation to be
     *   stored is known) */
    
    /*    public Table(Relation myPeer, int capacity) {
        peerRelation = myPeer;
        tuples = new Vector(capacity);
        }*/

    /** constructs a relation extent object with a default size. This
     *   is good for efficiency (if the size of the relation to be
     *   stored is known) */
    
    public Table(int capacity) {
        //        peerRelation = null;
        tuples = new Vector(capacity);
    }

    /** a copy constructor used to copy the extent. Necessary for
        comparing old and new view extent in the EVE-Demo */

    public Table(Table oldTable) {
        if (DEBUG) System.err.println("Copy constructor of MW.Table: Copying table "+oldTable.getRelName()+".");
        relName = new String(oldTable.getRelName());
        tuples = (Vector)oldTable.getTuples().clone();
        attributeNames = (Vector)oldTable.getAttributeNames().clone();
        attributeSizes = (Vector)oldTable.getAttributeSizes().clone();
    }        

    /** return this Table's vector of tuples (the reference,
        not a clone) */

    public Vector getTuples () {
        return tuples;
    }
    
    /** gets the tuples at index i
     * @param i the index of the tuple to be retrieved
     * @return null if the index is invalid */
    public Tuple getTuple(int i) {
        if (tuples == null) return null;
        if (i>=tuples.size()) return null;
        return (Tuple)tuples.elementAt(i);
    }

    /** returns the number of tuples in this relation.
     * @return -1 if there are no tuples
     */
    public int getSize() {
        if (tuples==null) return -1;
        return tuples.size();
    }

    /** return this Table's Relation reference */
    // private Relation getRelationx() {
    //   return peerRelation;
    // }

    /** adds a tuple at the end of this vector. If Vector is empty,
     *   creates a new vector with 10 original entries. Bag
     *   semantics. */
    public void addTuple (Tuple t) {
        if (tuples==null) tuples = new Vector (10);
        tuples.addElement(t);
    }

    /** searches for a certain tuple with the tuple's equals method
     *   and removes it from the extent. This is done
     *   <code>multiplicity</code> times.
     * @return the actual number of removed tuples. If this is less
     * than <code>multiplicity</code> of 0, then something is wrong */
    public int removeTuple (Tuple t, int multiplicity) {
        
        int removeCounter = 0;
        for (int j = 0 ; j < multiplicity ; j++) {
            for (int i = 0 ; i < tuples.size() ; i++ ) {
                if (i >= tuples.size()) break; 
                // since the vector may have become shorter in this loop;
                if (((Tuple)tuples.elementAt(i)).equals(t)) {
                    tuples.removeElementAt(i);
                    removeCounter++; // one tuple removed
                }
            }
        }
        return removeCounter;
    }

    /** string representation (calls toString for all tuples, one line each).
     * @return null if no tuples in the relation */
    public String toString() {
        String s = new String();
        if (tuples == null) return null;
        Enumeration e = tuples.elements();
        while (e.hasMoreElements()) {
            s += ((Tuple)e.nextElement()).toString()+"\n";
        }
        return s;
    }


    public String getRelName() {
        return relName;
    }
    
    public void setRelName(String name) {
        relName = name;
    }

    public Vector getAttributeNames() {
        return attributeNames;
    }

    public void setAttributeNames(Vector attNames) {
        attributeNames = attNames;
    }

    public Vector getAttributeSizes() {
        return attributeSizes;
    }

    public void setAttributeSizes(Vector attSizes) {
        attributeSizes = attSizes;
    }
}





