////////////////////////////////////////////////////////////////// COPYRIGHT //
//
//  EVE 2.0 - Evolvable View Environment
//
//  Copyright 1997-2000, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty. 
//  In no event will the authors be held liable for any damages arising 
//  from the use of this software.
// 
//  Authors:
//      Andreas Koeller, Xin Zhang, Elke Rundensteiner
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/dsrg/EVE
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to the
//  EVE Development Team: <rundenst@cs.wpi.edu>.
//
////////////////////////////////////////////////////////////////// COPYRIGHT //
package DSRG.MW;

import java.util.Vector;
import java.util.Enumeration;
import DSRG.MKB.*;

/** this class holds a tuple in a relation object. This is used for
 * storing (cacheing) relational extents in the EVE-Demo. This is
 * currently in the MKB, where it doesn't belong
@author koeller (04/07/1999) */


public class Tuple { 
    
    /** the values in the tuple. The types of the elements are
     *   heterogeneous, but have to agree with the types in the
     *   "attribute" objects referenced in the attributes-vector in
     *   this tuple's Table-object. */
    Vector values; 

    /** The extent to which this tuple belongs. Tuples belong to
     *   exactly one Table instance */
    Table parent;


    /** constructs a tuple with a certain Table instance as
     *   parent and sets its value from a vector (not cloned). */
    public Tuple (Table myParent, Vector myValues) {
        parent = myParent;
        values = myValues;
    }

    /** returns an enumeration on this tuple. The elements are of type
     *  "Object" */
    public Enumeration getEnumeration () {
        return values.elements();
    }

    /** returns the number of elements (attributes) in thi s
     * tuple. This should correspond to the number of attributes in
     * this tuple's relation extent's relation object, but this is not
     * enforced or checked... */
    public int getSize() {
        return values.size();
    }

    /** returns this tuple as a vector */
    public Vector getVector () {
        return values;
    }
    
    /** sets this tuple's vector (=data). Does not clone the
     *   contents. */
    public void setVector(Vector myValues) {
        values = myValues;
    }

    /** recursively calls the equals methods of this vector's objects */
    
    public boolean equals (Tuple otherTuple) {
        if (values.size() != otherTuple.getVector().size()) return false;
        Enumeration otherEnum = otherTuple.getEnumeration();
        Enumeration thisEnum = getEnumeration();
        while (otherEnum.hasMoreElements()) {
            if (!thisEnum.nextElement().equals(otherEnum.nextElement()))
                return false;
        }
        return true;
    }
            

    /** calls the toString() methods of this tuple's objects and
        concatenates them */
    public String toString () {
        String s= new String();
        for (int i=0; i<values.size(); i++) {
            s += values.elementAt(i).toString()+"|";
        }
        return s;
    }

    /** calls the toString() methods of this tuple's objects and
     *   concatenates them with commas in between and single quotes
     *   around the strings. This is used for storing this tuple in a
     *   JDBC database */
    public String toSQLString () {
        String s= new String();
        for (int i=0; i<values.size(); i++) {
            s += "'" + values.elementAt(i).toString()+"',";
        }
        return s.substring(0,s.length()-1); // cut last comma
    }    


    public synchronized Object clone() {
        Tuple tempTuple = new Tuple(parent,(Vector)values.clone());
        return tempTuple;
    }
}





