import java.util.*;
import java.io.*  ;
import DSRG.MKB.* ;
import DSRG.MW.*  ;
import INDModel.* ;

public class Discovery {

    public static  final boolean DEBUG = false;

    // use this number when no heuristics are being used
    private static final int DISABLED = 1000000;

    /** the maximal number of nodes retained */
    public static final int MAX_NUM_NODES = 200;

    /** the maximal number of edges in G_2 */
    public static final int MAX_NUM_EDGES = 3000;

    /** the maximum number of cliques actually tested against the DB
     * after a clique-generating algorithm finishes. */
    public static final int MAX_CLIQUES_TESTED =2000;

    /** maximum number of cliques retained for next step. For each of
     * those cliques, many sub_INDs will be generated, so the number
     * should be small. */
    public static final int MAX_CLIQUES_KEPT =1000;

    /**maximum number of INDs created for any one clique. That
     * restrict the maximum clique size that is successfully considered
     * for sub-INDs. Worsens the results but speeds up runtime. This
     * multiplied with MAX_CLIQUES_KEPT is the maximumnumber of INDs
     * generated at any one step in the algo. */
    public static final int MAXINDS_GENERATED=2000;
    

    /** the minimum number of distinct values necessary for the clique
     * to be considered "non-accidental". When setting both numbers to
     * 0, the heuristic is disabled.*/
    public static final int MIN_NUM_DISTINCT_UNARY = 7;
    public static final int MIN_NUM_DISTINCT_BINARY = MIN_NUM_DISTINCT_UNARY;
    
    /*============================================================
     *============================================================
     *============================================================
     *============================================================
     */

    Properties props;
    
    /**
     * a vector of DBManager objects to connect to physical databases 
     */
    Vector dBManagers = new Vector();

    
    /**
     * a handle (type DBManager) for the local DB-Manager
     */
    DBManager localDBManager;
 

    public int globalNumOfDBQueries=0;

    /** an array of Vectors. At the end of the algorithm, this will
        hold only the maximal cliques. In intermediate stages, it may
        also hold cliques that are later subsumed. Only the good
        cliques/INDs go here */
    private Vector resultINDs[];


    /** a reference to the MKB */
     
    MKB mainMKB;
      
    /** a reference to the checkIND class. Could be static, but that's
        bad Java */
    
    checkInd checkIndInstance;

    /** a file to record the INDs found. */

    String filename;
    PrintWriter outfile;
    


    public Discovery(String rel1, String rel2, String filenameP) {

	try{
	    props = new Properties(System.getProperties());
	    props.load(new BufferedInputStream
                (new FileInputStream("Discovery.properties")));
	    System.setProperties(props);
	}catch(Exception e){
	    e.printStackTrace();
	}

        filename = filenameP;

        init();

        // we have only one IS, so we might as well find the DBManager here.
        Relation tmpRel =mainMKB.findRelation("IS1",rel1 );
        DBManager tmpDBM = findDBManager(tmpRel.getIS());
        checkIndInstance = new checkInd(this, tmpDBM, rel1, rel2);
    
        System.out.println("Writing output to "+filename);
        System.out.println("Started at :"+System.currentTimeMillis()/1000);
        long starttime = System.currentTimeMillis();

        INDCollection myCollection = 
            new INDCollection( mainMKB.findRelation("IS1",rel1 ),
                               mainMKB.findRelation("IS1",rel2),
                               checkIndInstance
                               );
        int numINDs;
        Vector nodes=null; // ==unaryINDs
        Vector kAryINDs;
        TreeSet cliques=null;
        //Vector result=new Vector(); 
        Clique tmpClique; int tmpCliqueSize;
        boolean alreadySubsumed;                     
        Clique localClique; int localCSize; // for duplicate removal
        Ind tmpIND,tmpIND2,tmpIND3;
        boolean isSubsumed;
        
        // holds temporary candidates for cliques
        Vector tmpINDs;


        resultINDs = new Vector[myCollection.getNumAtts()+1];
        for (int k=1; k <= myCollection.getNumAtts(); k++)
            resultINDs[k]=new Vector();
        
        for (int k=1; k <= myCollection.getNumAtts(); k++) {

            //            int numInds = myCollection.generateAllInds();
            if (k==1) {
                nodes = myCollection.generateUnaryINDs();
                System.out.println(nodes.size()+ " INDs generated (1).");
                numINDs = myCollection.checkKAryInds(nodes);
                nodes = myCollection.removeInvalidINDs(nodes);
                System.out.println(nodes.size()+ " INDs validated (1).");
                System.out.println("Nodes are: "+nodes);
            }
            if (k==2)  {
                kAryINDs = myCollection.generateBinaryINDs(nodes);
                System.out.println(kAryINDs.size()+ " INDs generated (2)");
                numINDs = myCollection.checkKAryInds(kAryINDs);
                kAryINDs = myCollection.removeInvalidINDs(kAryINDs);
                System.out.println(kAryINDs.size()+ " INDs validated (2).");
                cliques = myCollection.generateCliquesAndCheck(nodes,kAryINDs,k);
                System.out.println(cliques.size()+" cliques checked and validated");
                if (DEBUG) System.out.println("The cliques for k=2 are: "+cliques);


                // now add preliminary 1- and 2-cliques.
                for (Iterator iter = cliques.iterator(); iter.hasNext();) {
                    tmpClique = (Clique)iter.next();
                    if(tmpClique.getCliqueSize()<=2)
                        resultINDs[tmpClique.getCliqueSize()].addElement(tmpClique.toIND(nodes));
                }

            }

            if (k>2 && cliques!=null) {
                // at this point, we have cliques generated from (k-1)
                // in vector "cliques", they are tested, so we can
                // check their state (since it's correct)

                System.out.println("Adding new cliques to result");
                //int counter = MAX_CLIQUES_KEPT; not necessary anymore
                for (Iterator iter=cliques.iterator(); iter.hasNext();) {
                    //if (counter--==0) break;
                    tmpClique = (Clique)iter.next();
                    tmpCliqueSize = tmpClique.getCliqueSize();
                    tmpIND = tmpClique.toIND(nodes);
                    // not sure why or even if we generate multiple cliques here
                    isSubsumed=false;
                    for(int i=tmpCliqueSize; i<=myCollection.getNumAtts() && !isSubsumed; i++) {
                        for (Iterator iter2=resultINDs[i].iterator(); iter2.hasNext();)
                            if (((Ind)iter2.next()).contains(tmpIND)) {
                                isSubsumed=true;
                                break;
                            }
                    }
                        
                    if (!isSubsumed) {
                        if (DEBUG) System.out.println("Trying to add "+tmpClique);
                        if (tmpClique.getState()==true && tmpClique.getCliqueSize()>=k-1 /*&& !resultINDs[tmpCliqueSize].contains(tmpClique.toIND(nodes))*/)
                            resultINDs[tmpClique.getCliqueSize()].addElement(tmpClique.toIND(nodes));
                    
                    }
                    else if (DEBUG) System.out.println(tmpClique +"not added since already subsumed.");
                    
                }

                // now generate INDs from falsified cliques, check
                // those INDs and generate new (hyper-)cliques
                kAryINDs = myCollection.generateIndsFromCliques(k,nodes,cliques);
                System.out.println(kAryINDs.size()+ " INDs generated (3).");
                //                if (DEBUG) System.out.println("Those INDs are: "+kAryINDs);
                if (kAryINDs.size()==0) break; // if no new INDs, we're done.

                numINDs = myCollection.checkKAryInds(kAryINDs);

                System.out.println(numINDs+ " INDs validated (3).");

                // now we have all the k-ary cliques (tested). We need
                // to generate all maximal (k-1)-ary INDs since we
                // won't discover them otherwise


                // this time, we need the invalid INDs first.
                tmpINDs = new Vector();
                for (Iterator iter = kAryINDs.iterator(); iter.hasNext();) {
                    tmpIND=(Ind)iter.next();
                    // for each false IND
                    if (tmpIND.getState()==Ind.GREY) throw new RuntimeException("found a grey IND here... something's wrong: "+tmpIND);
                    if (tmpIND.getState()==Ind.BLACK) {
                        // generate all k-1-subINDs
                        for(int i =0; i<tmpIND.getArity(); i++) {
                            tmpIND2=(Ind)tmpIND.removeIndAt(i);
                            // if one of them is NOT subsumed by a true IND
                            isSubsumed=false;
                            for(Iterator iter2=kAryINDs.iterator(); iter2.hasNext();) {
                                tmpIND3=(Ind)iter2.next();
                                if (tmpIND3.getState()==Ind.WHITE && tmpIND3.contains(tmpIND2)) {
                                    isSubsumed=true;
                                    break;
                                }
                            }
                            // Question: is it possible here that a
                            // tmpIND2 is subsumed by an old
                            // (previously discovered) clique but NOT
                            // by any true IND and by none of the
                            // tmpINDs? I'm not sure (as of 071801)
                            if (!isSubsumed && !tmpINDs.contains(tmpIND2)) tmpINDs.addElement(tmpIND2);
                        }
                    }
                }                

                if (DEBUG) System.out.println(tmpINDs.size()+" (k-1)-ary maximal INDs generated from falsified cliques");
                // now add these new cliques to the final solution
                for (Iterator iter=tmpINDs.iterator(); iter.hasNext();) {
                    tmpIND = (Ind)iter.next();
                    if (tmpIND.getArity()!=k-1) throw new RuntimeException("Something is wrong here. Ind "+tmpIND+" has the wrong size, should be "+(k-1));
                    isSubsumed=false;
                    for(int i=k; i<=myCollection.getNumAtts() && !isSubsumed; i++) {
                        for (Iterator iter2=resultINDs[i].iterator(); iter2.hasNext();)
                            if (((Ind)iter2.next()).contains(tmpIND)) {
                                isSubsumed=true;
                                break;
                            }
                    }
                    if (!isSubsumed) {
                        if (DEBUG) System.out.println("Trying to add "+tmpIND+ " to final result.");
                        resultINDs[k-1].addElement(tmpIND);
                        
                    }
                    else if (DEBUG) System.out.println(tmpIND +"not added since already subsumed.");
                }


                kAryINDs = myCollection.removeInvalidINDs(kAryINDs);
                System.out.println("Removed invalid INDs, remaining INDs: "
                                   +kAryINDs.size());
                if (kAryINDs.size()==0) break; // if no new INDs, we're done.

                System.out.println("Generating new cliques for level "+k);
                cliques = myCollection.generateCliquesAndCheck(nodes,kAryINDs,k);
                System.out.println("Generated "+cliques.size()+" new cliques.");
                if (DEBUG) System.out.println("Cliques are: "+cliques);
            }

            
            System.out.print("Cliques for k="+k+": ");
            for (int i = 1; i < resultINDs.length; i++) 
                System.out.print(resultINDs[i].size()+" ");
            System.out.println();

            //if (k>=2 && cliques==null) break; // we're done
        }
        
        outfile.println("System parameters:"+
                        "\nMAX_NUM_NODES="+MAX_NUM_NODES+
                        "\nMAX_NUM_EDGES="+MAX_NUM_EDGES+
                        "\nMAX_CLIQUES_TESTED="+MAX_CLIQUES_TESTED+
                        "\nMAX_CLIQUES_KEPT="+MAX_CLIQUES_KEPT+
                        "\nMAXINDS_GENERATED="+MAXINDS_GENERATED+
                        "\nMIN_NUM_DISTINCT_UNARY="+MIN_NUM_DISTINCT_UNARY+
                        "\nMIN_NUM_DISTINCT_BINARY="+MIN_NUM_DISTINCT_BINARY
                        );
        
        System.out.println("The final INDs are stored in  "+filename);
        for (int i=1; i<resultINDs.length; i++)
            outfile.println(i+": "+resultINDs[i].toString());

        // IND-Merging

        System.out.println("Trying to merge INDs.");

        Ind testInd=null,backupInd=null,tmpInd;

        boolean stillTrue = true;
        Iterator iter;
        // cycling down, starting from larger INDs, keep merging until
        // IND does no longer hold.
        for (int i=resultINDs.length-1; i>0 && stillTrue; i--) {
            if (((Vector)resultINDs[i]).size()==0) continue;
            // now i has the highest index for an IND
            iter = ((Vector)resultINDs[i]).iterator();
            if (testInd==null) {
                testInd=(Ind)iter.next();
                backupInd = testInd;
            }
            while(iter.hasNext() && stillTrue) {
                backupInd = testInd;
                tmpInd = (Ind)iter.next();
                testInd = testInd.merge(tmpInd);
                if (checkIndInstance.Main(testInd)<0) stillTrue=false;
            }
        }
        
        outfile.print("Largest valid merged IND (Size ");
        if (stillTrue) 
            outfile.println(testInd.getArity()+"): "+testInd);
        else
            outfile.println(backupInd.getArity()+"): "+backupInd);
        //outfile.println("Smallest invalid merged IND: "+testInd);

        outfile.println("Number of INDs discovered:");
        int totalNumber =0;
        for (int i=1; i<resultINDs.length; i++) {
            outfile.println(i+ ": "+((Vector)resultINDs[i]).size());
            totalNumber+=((Vector)resultINDs[i]).size();
        }
        outfile.println("Total: "+totalNumber);
        
        
        System.out.println("Finished at :"+System.currentTimeMillis()/1000);
        System.out.println("Runtime: "+(System.currentTimeMillis()-starttime)/1000+" seconds");

        outfile.println("Number of DB queries: "+globalNumOfDBQueries);
        outfile.flush();
        outfile.close();

    }


    /**
     * Initialize MKB. <BR>The database is not being reset here,
     * this is done with the separate class "resetDB" from a shell
     * script (once implemented). */
    public void init() {

        //        queryEngine = new QueryEngine(this);

        try {
        outfile = 
            new PrintWriter (new FileOutputStream (filename));

        }
        catch (Exception e) {e.printStackTrace();}

	String mkbName = System.getProperty("mkbFileLocation");
	mainMKB=new DSRG.MKB.MKB(mkbName);
	if (mainMKB == null)
	    System.err.println("MKB not initialized");
        
        
        Enumeration ISs = mainMKB.getISs().elements();
        IS oneIS;

        // create one DBManager for each IS
        while (ISs.hasMoreElements()) {
            oneIS = (IS)ISs.nextElement();
            if (!oneIS.getName().equals("LOCAL_IS")) {
                if (DEBUG) System.err.println("Creating DBManager for "+oneIS.ISName);
                
                dBManagers.addElement(new DBMOracle(oneIS.ISName,
                                                    oneIS.username,
                                                    oneIS.password,
                                                    oneIS.DBURL));
            }
        }
        
        // and the DBManager for the local IS.
        oneIS = mainMKB.getLocalIS();
        localDBManager = new DBMOracle(oneIS.ISName,
                                       oneIS.username,
                                       oneIS.password,
                                       oneIS.DBURL
                                       );

        
    }

    public Vector getISs() {
        return mainMKB.getISs();
    }

    /** finds the DBManager connected to an IS */
    public DBManager findDBManager(IS is) {
        DBManager tmpDBM;
        for (Iterator iter=dBManagers.iterator();iter.hasNext();) {
            tmpDBM = (DBManager)iter.next(); 
            if ( tmpDBM.getISName().equals(is.getName()))
                return tmpDBM;
        }
        return null;
    }
}
