package INDModel;

import java.sql.*;
import java.util.*;
import java.io.*;
import Discovery;
import DSRG.MW.DBManager;

/** when this class is instantiated, it computes the value
    distribution histograms of all attributes of two relations (passed
    in the constructor). Then, function isRelated(attname1, attname2)
    will return a boolean value that is true when the two attribute
    names are refering to related attributes 
*/
class ChiSquare {


    private  boolean DEBUG=false;

    Vector hist1=null, hist2=null;
    Vector colNames1=new Vector(), colNames2=new Vector();
    
    Discovery myMainClass;
    DBManager tmpDBM;


    public  ChiSquare(Discovery mainClass, DBManager dBM, String relname1, String relname2) {

        myMainClass = mainClass;

        // get DBManager from Hashtable
        tmpDBM = dBM;

        hist1 = getHistograms(tmpDBM.connection,relname1,colNames1);
        hist2 = getHistograms(tmpDBM.connection,relname2,colNames2);
        
    }


    /** this function returns the QUOTIENT between the Percentile for
        95% confidence and the Chi-Square-Value. That is, a low number
        (<1) means the attributes are likely unrelated, a high number
        means we don't know.  An exact zero means (since there are no
        zero-percentiles) that we don't have a histogram for this
        pairing.  */
    double isRelated(String attname1, String attname2) {

        // We now have two vectors, with hashtables in each vector that
        // represent histograms of value distributions in each table.
        
        // So we now need to compute the chi-square values of each
        // pair of attributes here

        //double chis[][] = new double[hist1.size()][hist2.size()];
        //int df[][] = new int[hist1.size()][hist2.size()]; // degrees of freedom

        if (DEBUG) System.out.println("Checking "+attname1+" vs. "+attname2);

        int i=colNames1.indexOf(attname1);
        int j=colNames2.indexOf(attname2);

        // for empty attributes (no histogram)
        if (i==-1 || j==-1) {
            if (DEBUG) System.out.println("Attribute not found in colnames");
            return 0; 
        }
            
        /*
        for (int i=0; i< hist1.size(); i++) {
            if (!colNames1.elementAt(i).equals(attname1)) continue;
            for (int j=0; j<hist2.size(); j++) {
                if (!colNames2.elementAt(j).equals(attname2)) continue; */
                // For that, we take a pair of attributes, one from each
                // Vector, and first compute the union of hash values. In
                // order to do that, we create two more hashtables from the
                // current ones, and crosswise add attributes that don't exist
                // in the other hashtable.

                if (DEBUG) System.out.println(colNames1.elementAt(i)+" "+colNames2.elementAt(j));
        
                Hashtable h1 = (Hashtable)hist1.elementAt(i);
                Hashtable h2 = (Hashtable)hist2.elementAt(j);

                if (DEBUG) System.out.println("Original Hashtables: "+h1+"\n"+h2);
                Hashtable h1full = (Hashtable)h1.clone();
                Hashtable h2full = (Hashtable)h2.clone();

                String temp;

                for (Enumeration enum=h1.keys(); enum.hasMoreElements();) {
                    temp = (String)enum.nextElement();
                    if (!h2full.containsKey(temp)) {
                        h2full.put(temp,new Integer(0));
                    }
                }

                for (Enumeration enum=h2.keys(); enum.hasMoreElements();) {
                    temp = (String)enum.nextElement();
                    if (!h1full.containsKey(temp)) {
                        h1full.put(temp,new Integer(0));
                    }
                }

                if (DEBUG) {
                    System.out.println("Padded Hashtables: ");
                    System.out.println(h1full);
                    System.out.println(h2full);
                }

                // Now we store the two Hashtables in an array, then
                // compute the row and column sums and the matrix of
                // expected values.

                int cols = h1full.size();
                int O[][] = new int[2][cols]; // observations
                double E[][] = new double[2][cols]; // expectations
                
                int col=0;
                // careful here, we need to fill the array in the same
                // order as the hashtable for the first array
                // suggested!
                String hashKey;
                for (Enumeration enum=h1full.keys(); enum.hasMoreElements();) {
                    hashKey = (String)enum.nextElement();
                    O[0][col]=((Integer)h1full.get(hashKey)).intValue();
                    O[1][col]=((Integer)h2full.get(hashKey)).intValue();
                    col++;
                }

                if (DEBUG) {
                    System.out.println("O-Array:");
                    for (int r=0;r<2;r++) {
                        for (int c=0;c<cols;c++) {
                            System.out.print(O[r][c]+" ");
                        }
                        System.out.println();
                    }
                }

                int rowsums[] = new int[2];
                int colsums[] = new int[cols];

                // compute row sums
                int sum;
                for(int r=0; r<2; r++) {
                    sum =0;
                    for (int c=0; c<cols; c++) {
                        sum+=O[r][c];
                    }
                    rowsums[r]=sum;
                    if (DEBUG) System.out.println("Rowsum "+r+"= "+sum);
                }
                
                // compute column sums
                for (int c=0; c<cols; c++) {
                    colsums[c]=O[0][c]+O[1][c];
                    if (DEBUG) System.out.println("Colsum "+c+"= "+colsums[c]);
                }


                double total = (double)(rowsums[0]+rowsums[1]); // the
 // grand total, as double for precision of the E-table
                if (DEBUG) System.out.println("Total # of observations: "+total);
                
                // compute E
                for (int r=0; r<2; r++) {
                    for (int c=0; c<cols; c++) {
                        E[r][c]=rowsums[r]/total*colsums[c]; // overflow problem
                        if (DEBUG) System.out.print(E[r][c]+" ");
                    }
                    if (DEBUG) System.out.println();
                }
                

                // now compute \chi^2=\sum{\frac{(O-E)^2}{E}}

                double chisq =0.0;
                
                for (int r=0;r<2; r++)
                    for (int c=0;c<cols;c++) {
                        chisq+=sqr(O[r][c]-E[r][c])/E[r][c];
                        if (DEBUG) System.out.print(chisq+" ");
                    }

                if (DEBUG) System.out.println("chi-square("+colNames1.elementAt(i)+", "+colNames2.elementAt(j)+") = "+chisq);

                
                /*
                chis[i][j]=chisq;
                df[i][j]=cols-1; // times r-1, which is one
            }
        }
                */
                /*
        if (!DEBUG) {
            for(int j=0; j<hist2.size(); j++) { // columns
                System.out.print(colNames2.elementAt(j)+" ");
            }
            System.out.println();
            
            for(int i=0; i<hist1.size(); i++) {
                System.out.print(colNames1.elementAt(i)+" ");
                for(int j=0; j<hist2.size(); j++) {
                    //System.out.print(chis[i][j]+" ");
                    if (chis[i][j]<chisq_quantile(df[i][j])) 
                        System.out.print("1 ");
                    else
                        System.out.print("0 ");
                }
                System.out.println();
            }
        }
                */
                // degrees of freedom are (cols-1)(rows-1)=cols-1
                //return (chisq<chisq_quantile(cols-1)?true:false);
                return chisq_quantile(cols-1)/chisq;
    }

     double sqr(double a) { return a*a; }
    

    /** this function computes the histograms for a table and returns
     a vector of them. Vector colNames is also a return parameter (all
     names that have histograms) and should thus be initialized with
     <code>=new Vector()</code>. Each Histogram is a hashtable of
     values and counts. This algorithm will produce garbage if the
     attribute whose hashtable is computed is a BLOB or anything else
     that doesn't fit into a hashtable.  <P>[082401] Only histograms
     of attributes with less than MAX_NUM_DISTINCT distinct values
     will be computed and returned.
     @return Vector of Histograms.  */
    
    Vector getHistograms(Connection con, String tableName, Vector colNames) {

        Vector result = new Vector();
        Hashtable hist;


        Statement stmt=null;
        ResultSet rs=null;

        try {

            stmt =con.createStatement(); 
            rs = stmt.executeQuery("select column_name from user_tab_columns where table_name ='"+tableName.toUpperCase()+"'");
            
            while(rs.next()) {
                colNames.addElement(rs.getString(1));
            }
        }
        catch (SQLException e) {e.printStackTrace();}

        String a;
        String value;
        int count=-1;
        
            
        for (int i = 0 ; i< colNames.size(); i++) {
            System.out.print("H"); // progress marker
            a=(String)colNames.elementAt(i);

            try {
                rs=stmt.executeQuery("select count(*) from (select distinct "+
                                     a+
                                     " from "+
                                     tableName+
                                     ")"
                                     );
                if (!rs.next()) System.err.println("Something's weird, we got no tuples from a count-query");
                count = rs.getInt(1);

            }
            catch (SQLException e) {e.printStackTrace();}

            if (count>Discovery.MIN_NUM_DISTINCT_UNARY) {
                    colNames.removeElement(a);
                    i--; //old stupid for-loop-remove-from-vector-trick
            }
            else {
                hist = new Hashtable();
                try {
                    rs = stmt.executeQuery("select "+
                                           a+
                                           ", count( "+
                                           a+
                                           ") from "+
                                           tableName+
                                           " group by "+
                                           a);
                    
                    while (rs.next()) {
                        value = rs.getString(1);
                        count = rs.getInt(2);
                        if (value!=null) {
                            hist.put(value,new Integer(count));
                        }
                    }
                    if (hist.size()>0) 
                        result.addElement(hist);
                    else {
                        colNames.removeElement(a);
                        i--; //old stupid for-loop-remove-from-vector-trick
                    }
                }
                catch (SQLException ex) {
                    // incompatible attribute types, so no IND
                    System.out.println("SQL Exception at "+i+":\n"+ex.toString());
                }
            }
        }
        // close the statement

        try {
            stmt.close();
        }
        catch (SQLException e) {e.printStackTrace();}

        System.out.println();
        return result;
    }

    /** this function returns the quantile of the chi-square
     distribution for alpha=0.95, i.e., it returns, for degree of
     freedom df, which chi-square-value has to be exceeded to refute
     the hypothesis that the input variables are related.  A typical
     relative error of this function between "df"s of 41 and 69 is
     10%, which however should suffice for this purpose.
     @param df Degrees of Freedom */
     double chisq_quantile(int df) {
        double[] chi69 = { 
        1e99,  3.84, 5.99, 7.81, 9.49,11.1, 12.6, 14.1, 15.5, 16.9,
        18.3, 19.7, 21.0, 22.4, 23.7, 25.0, 26.3, 27.6, 28.9, 30.1,
        31.4, 32.7, 33.9, 35.2, 36.4, 37.7, 38.9, 40.1, 41.3, 42.6,
        43.8, 43.8, 43.8, 43.8, 43.8, 43.8, 43.8, 43.8, 43.8, 43.8, 
        55.8, 55.8, 55.8, 55.8, 55.8, 55.8, 55.8, 55.8, 55.8, 55.8, 
        67.5, 67.5, 67.5, 67.5, 67.5, 67.5, 67.5, 67.5, 67.5, 67.5,
        79.1, 79.1, 79.1, 79.1, 79.1, 79.1, 79.1, 79.1, 79.1, 79.1
        };
        
        double normal95=1.644853626;
        int cutoff = 70;
        
        if (df<cutoff) 
            return chi69[df];
        else 
            return (.5*sqr(normal95+Math.sqrt(2*df-1)));
    }

} 
