package INDModel;

import java.util.*;
import Discovery;
import DSRG.MKB.*;
import DSRG.MW.*;
import java.sql.*;

/** this class will implement the discovery algorithm on the relation
level. It receives and IND and checks it. The method by which this is
done is to be determined. */

public class checkInd {
    /** this contains a list of DBManagers, indexed by the IS names
        used here */

    private final static boolean DEBUG = false;

    Hashtable dbManagers = new Hashtable();

    Discovery myMainClass;

    ChiSquare myChiSquare;
    /** the handle to the main class is needed to associate DBManagers
        with ISs. For speedup, this is a hashtable that is initialized
        only once. */
    public checkInd(Discovery mainClass,DBManager DBM 
                    /* hack for single-DBM systems only*/, String lRel, String rRel) {
        IS tmpIS;
        for(Iterator iter = mainClass.getISs().iterator();
            iter.hasNext();) {
            tmpIS = (IS)iter.next();
            dbManagers.put(tmpIS,mainClass.findDBManager(tmpIS));
        }
        myMainClass = mainClass;
        
        myChiSquare = new ChiSquare(mainClass, DBM, lRel, rRel);
        
    }

    /** checks if an IND is valid in the database. Returns a positive
        number (the "support" if it is valid, and a negative number if
        it is not. */
    public double Main(Ind ind) {
        // ok, let's check this stuff against the database. For stage
        // one, we assume that the two relations are in the same DB so
        // can be queried against each other with a MINUS query. So
        // we'll just construct MINUS-queries containing the two
        // relation names and the projection lists.

        if (DEBUG) System.out.println("checking "+ind.toString());

        if (ind.getState()==Ind.WHITE) System.out.println("don't need to test. IND already white: "+ind);

        // get DBManager from Hashtable
        DBManager tmpDBM = (DBManager)dbManagers.get(ind.lRel.getIS());

        String lAttList = new String();
        String rAttList = new String();
        Attribute tmpAtt;

        // num of distinct values in first attribute
        int firstCount=0, secondCount=0;
        // ratio between firstcount and secondcount
        double ratio=0;
        
        if (ind.getArity()==1) {// type incompatibilities possible
            String leftType = ((Attribute)ind.lAtts.elementAt(0)).getType();  
            String rightType = ((Attribute)ind.rAtts.elementAt(0)).getType();
            if (!leftType.equals(rightType)) {
                System.out.print("*");
                return -1.0;
            }
        }

        myMainClass.globalNumOfDBQueries++;
        
        for (Iterator iter = ind.lAtts.iterator();
             iter.hasNext();) {
            tmpAtt = (Attribute)iter.next();
            lAttList += tmpAtt.getName();
            if (iter.hasNext()) lAttList +=",";
        }

        for (Iterator iter = ind.rAtts.iterator();
             iter.hasNext();) {
            tmpAtt = (Attribute)iter.next();
            rAttList += tmpAtt.getName();
            if (iter.hasNext()) rAttList +=",";
        }

        
        // now form query.
        String query = "select count(*) from (select "+
            lAttList+
            " from "+
            ind.lRel.getName()+
            " minus select "+
            rAttList+
            " from "+
            ind.rRel.getName() +
            ")";
        
        // we need a simpler query mechanism here, so we make one.
        ResultSet rs=null;
        int indCount=-1;
        boolean INDvalid=false;
        try {
            Statement stmt = tmpDBM.connection.createStatement();
            
            //        DSRG.MW.Table table = tmpDBM.connection.executeQuery(query);
            
            rs = stmt.executeQuery(query);
            if (!rs.next()) System.err.println("Something's weird, we got no tuples from a count-query");
            indCount = rs.getInt(1);
            if (indCount==0) INDvalid = true;
            stmt.close();
            
        }
        catch (SQLException ex) {
            INDvalid = false;
            // ORA-01790 is "Incompatible Datatypes", meaning
            // attribute types did not match, i.e., there can be no IND
            if (ex.toString().indexOf("ORA-01790")==-1) 
                System.out.println("Exception in DB: "+ex.toString());
            else System.out.println("Incompatible Types: "+((Attribute)ind.lAtts.elementAt(0)).getType()+", "+((Attribute)ind.rAtts.elementAt(0)).getType());
            // wait one second
            long starttime = System.currentTimeMillis();
            while ((System.currentTimeMillis()-starttime)<1000);
        }


        // 072601 we run another experiment now. when an IND tests
        // valid, we check again and see if the distinct count of the
        // first attset is less than 1/10 of the distinct count of the
        // second one. If that is the case, the IND is considered
        // spurious (and thus false)

        // we generate queries like this one:
        //  select count(*) from ( select distinct child03,mailcode from cup98_40_2)      
        
        double chisq_ratio=-1;

        if (INDvalid) {
            // now form query.

            rs=null;
            try {
                query = "select count(*) from (select distinct "+
                    lAttList+
                    " from "+
                    ind.lRel.getName()+
                    ")";            
                
                Statement stmt = tmpDBM.connection.createStatement();
                
                
                rs = stmt.executeQuery(query);
                if (!rs.next()) System.err.println("Something's weird, we got no tuples from a count-query");
                firstCount = rs.getInt(1);
                
                query = "select count(*) from (select distinct "+
                    rAttList+
                    " from "+
                    ind.rRel.getName()+
                    ")";            
                
                rs = stmt.executeQuery(query);
                if (!rs.next()) System.err.println("Something's weird, we got no tuples from a count-query");
                secondCount = rs.getInt(1);
                stmt.close();
                
                ratio = (double)firstCount/(double)secondCount;
            }
            catch (SQLException ex) {
                INDvalid = false;
                // ORA-01790 is "Incompatible Datatypes", meaning
                // attribute types did not match, i.e., there can be no IND
                if (ex.toString().indexOf("ORA-01790")==-1) 
                    System.out.println("Exception in DB: "+ex.toString()+". Query was: "+query);
                else System.out.println("Incompatible Types: "+((Attribute)ind.lAtts.elementAt(0)).getType()+", "+((Attribute)ind.rAtts.elementAt(0)).getType());
            }

            // fuzzy INDs: hardcoded  for INTERNET2/INTERNET99
            // if (indCount<(double)firstCount/9109.0*101.0) INDvalid=true;

            /*
            if (ratio<(firstCount>40?0.5:0.9-(double)firstCount/100.0)) {
                System.out.print("/");
                INDvalid=false;
            }
            */
            int k = ind.lAtts.size();
            if (k==1) {
                if (firstCount<=Discovery.MIN_NUM_DISTINCT_UNARY) {
                    String lAttName = ((Attribute)ind.lAtts.elementAt(0)).getName();
                    String rAttName = ((Attribute)ind.rAtts.elementAt(0)).getName();
                    
                    // optimistic approach: IND is only rejected if < 10
                    // distinct tuples in left attribute set, there is
                    // only one attribute, AND the attribute distributions
                    // do not match.
                    

                    chisq_ratio = myChiSquare.isRelated(lAttName,rAttName);

                    /* =======================================
                       as a test solely for Carolina, I'll disable this
                       heuristic
                    */
                    //                    chisq_ratio=1; // never reject
                    
                    if (!(chisq_ratio>.1)) { // looser heuristic, somewhat unscientific
                        System.out.print("@");
                        INDvalid =false;
                    }
                    else System.out.print("o");
                }
                else System.out.print("X");
            }
            
            if (k==2) {
                if (firstCount<=Discovery.MIN_NUM_DISTINCT_BINARY) {
                    System.out.print("2");
                    INDvalid=false;
                }
                else System.out.print("o");
            }
            if  (k>=3) System.out.print("X");
            
            
        }
        else System.out.print(".");
        
            
        if (INDvalid) {
            if (chisq_ratio<0)
                return firstCount*ratio; // no chisq was computed
            else
                return firstCount*ratio*Math.sqrt(chisq_ratio);
        }
        else return -2.0;
    }

    /* @return true if strings are considered similar
     */
    private static boolean similarity(String s1, String s2) {
        Hashtable h=new Hashtable();
        String trigram;
        for (int i=0; i<s1.length()-2; i++) {
            trigram=s1.substring(i,i+3);
            if (h.containsKey(trigram))
                h.put(trigram,new Integer(((Integer)h.get(trigram)).intValue()+1));
            else
                h.put(trigram,new Integer(1));
        }
        for (int i=0; i<s2.length()-2; i++) {
            trigram=s2.substring(i,i+3);
            if (h.containsKey(trigram))
                h.put(trigram,new Integer(((Integer)h.get(trigram)).intValue()-1));
            else
                h.put(trigram,new Integer(-1));
        }
        
        /* Hylton's
        int sumSq =0;
        for (Enumeration enum = h.elements(); enum.hasMoreElements();)
            sumSq += sqr(((Integer)enum.nextElement()).intValue());
        double score = Math.sqrt((double)sumSq);
        */

         //Ukkonen's
        int sum =0;
        for (Enumeration enum = h.elements(); enum.hasMoreElements();)
            sum += Math.abs(((Integer)enum.nextElement()).intValue());
        double score = (double)sum;
          
         // now using Hylton's empirical formula
        //       if (score <= 2.486 + 0.025*(h.size()))

        // Ukkonens'
        if (score<=0.6*(s1.length()+s2.length()-4))
            return true;
        else
            return false;

    }

    private static int sqr(int a) { return a*a;}

}

