package INDModel;

/** <PRE> Author: Ram Samudrala (me@ram.org)
 *         extend() was implemented from the Bron and Kerbosch
 *         algorithm (Communications of the ACM 16:575-577, 1973) by
 *         Jean-Francios Gibrat.
 * 
 * April 19, 1995.
 *<PRE>
 * Translated to Java in July 2001, Andreas Koeller. Debugged
 * sources. All older versions were buggy.
 * 
 * <P> This version is adapted as a class file to fit into my V1.0
 * discovery program.  <P> 071101 I'm about to add a
 * hyperclique-finding-algorithm that I wrote, to take care of the
 * higher steps in the process.  */

import java.util.*;
import java.io.*;


class findCliques {

    static TreeSet cliques;
    

    //    static final int MAX_CLIQUES=100000;
    static final int MAX_NODES=400; 
    static final boolean DEBUG = false;
    static String prefix=new String(); // for debugging only 


    static boolean[][] edge;

    static int cliqueIndex, nel, compsub[]= new int[MAX_NODES];
    /*
    static int clique_size(int x) {
        return ((Clique)cliques.elementAt(x)).getCliqueSize();
    }

    static int clique_vertex(int x, int i)  {
        return ((Clique)cliques.elementAt(x)).getCliqueVertex(i);
    }
    
    static void addCliqueVertex(int x, int vertex) {
        ((Clique)cliques.elementAt(x)).addCliqueVertex(vertex);
    }
    */
    /** finds cliques in a graph. Graph must be 0-based now (!) */

    static TreeSet findCliques(boolean graph[][]) {
        
        edge=graph;
        
        
        cliques = new TreeSet();
        
        
        int all[], nn = edge.length; // assuming graph is a quadratic
        // array. We moved to zero-based arrays now
        
        
        all = new int[nn];
        
        for(nel = 0; nel < nn; nel++)
            all[nel] = nel;
        nel = 0;	
        
        extend(all, 0, nn);
        
        return cliques;
    }

    /******************************************************************/

    static boolean extend(int old[], int ne, int ce) {
        /*
         * Author: Jean-Francois Gibrat
         *
         * This subroutine finds the maximal complete subgraphs (cliques) of a
         * given undirected graph. The algorithm consists in a recursively defined
         * extension operator (extend) applied to the three following sets:

         * set compsub: set to be extended by a new point or shrunk by one
         *              point as the algorithm travels along a branch of
         *              the backtracking tree.
         * set cand: set of all vertices connected to the current vertex
         *              in compsub i.e., vertices which will be used as an
         *              extension of the present configuration of compsub.
         * set not: set of all points which have already served as an
         *              extension and are now excluded. Or: set of nodes
         *              previously included in cliques that are CONNECTED
         *              to the selected node in the next higher stack
         *              frame
         *
         * The basic mechanism consists in the following steps:
         * Step 1:  selection of a candidate.
         * Step 2:  creating new sets cand and not from the old sets by removing all points
         *          not connected to the selected candidate, keeping the old sets intact.
         * Step 3:  adding the selected candidate to compsub.
         * Step 4:  calling the extension operator to operate on the sets just formed.
         * Step 5:  upon return the selected candidate is transfered from compsub to the 
         *          old set not.
         *
         * The set compsub is a stack external to the subroutine.
         * The sets not and cand are stored in the local array old,
         * not from positions 1 to ne and cand from positions ne+1 to ce.
         * nvert x nvert array [connected] indicates which vertices are connected to which
         * vertices. This array is external to the subroutine.
         * 
         * Last modification: 13/07/94
         *
         * Reference: Bron C. and Kerbosch J., Commun. A.C.M., vol. 16, pp 575-577, 1973.
         * */

	if (DEBUG) {
            prefix=prefix+"  ";
	    System.out.print(prefix+"Recursion.\n"+prefix+"Parameters: old=[");
	    for(int ii=0;ii<old.length; ii++)
		System.out.print(old[ii]+",");
	    System.out.println("], ne="+ne+", ce="+ce);
	}

	int newSets[]=new int[ce];  //= array [0:ce-1] of int
	int knod, nod, fixp;
	int newne, newce, i, j, count, pos, p, s, sel, minnod;


	// try to break it to find bugs:
	for (i=0; i<newSets.length;i++)
	    newSets[i]=-1;
	minnod = ce;//or Integer.MAX_VALUE;
	nod = 0; pos=-1;s=-1;fixp=-1;

	/* 
	 * Look for the vertex that is connected to the maximum number of vertices, i.e.,
	 * the minimum number of disconnections (Step 1)
	 *
	 */
	for(i = 0; (i < ce) && (minnod != 0); i++) {
            p = old[i];
            count = 0;

            pos = -1; // try to break it
      
            /* Count disconnections */
            for(j = ne; (j < ce) && (count < minnod); j++) {
                if(!edge[p][old[j]]) {
                    count++;
                    pos = j;  /* save position of potential candidate */
                }
            }
            /* Test whether vertex at position i in the stack has
               a minimum number of diconnections */
            if(count < minnod) {
                fixp = p;
                minnod = count;
                if(i < ne) {
                    s = pos;
                    nod=0;
                }
                else {
                    s = i;
                    nod = 1;  
                                /* Number of disconnections preincreased by one */
                }
            }
        }
  
	/* 
	 * Backtrack Cycle: the loop is executed as many times as
	 * there are vertices not connected to the selected point
	 *  */
  
        if (DEBUG) System.out.println(prefix+"fixpoint is "+fixp);
  
	for(knod = minnod + nod; knod >= 1; knod--) {
            if (DEBUG) System.out.println(prefix+"For loop. knod= "+knod);
            /* 
             * Interchange 
             */
            p = old[s];
            old[s] = old[ne];
            old[ne] = p;
            sel = old[ne];
            /* 
             * Fill new set not (Step 2)
             */
            newne = 0;
            for(i = 0; i < ne; i++)
                if(edge[sel][old[i]]) 
                    newSets[newne++] = old[i];	
            /* 
             * Fill new set cand (Step 2)
             */
            newce = newne;
            for(i = ne + 1; i < ce; i++)
                if(edge[sel][old[i]]) 
                    newSets[newce++] = old[i];

            /* 
             * The selected vertex is added to compsub (Step 3)
             */
            compsub[nel++] = sel; /* Note that index nel is
                                     external to the subroutine */
            /* 
             * Set cand and not are empty, a new clique has been
             * found: the recursive process terminates when this
             * condition is met.  
             * */

            if(newce == 0) { 
                /* The new clique is stored in compsub. */
                        
                if (DEBUG) {
                    System.out.print("Bron/Kerbosch found a clique: ");
                    for (int ii=0; ii<nel; ii++)
                        System.out.print(compsub[ii]+" ");
                    System.out.println();
                }
                Clique tmpClique = new Clique(nel);
                //                cliques.add(new Clique(nel)); // nel is the size
                //                cliqueIndex = cliques.size()-1;
                        
                for(i = 0; i < nel; i++) {
                    tmpClique.addCliqueVertex(compsub[i]);
                    //            addCliqueVertex(cliqueIndex, compsub[i]); 
                }
                if (DEBUG) {          
                    System.out.println(tmpClique);
                }
                cliques.add(tmpClique);
            }	
            /*
             * Otherwise calls the extension operator to operate
             * on the sets just formed (Step 4) */
            else { 
                if (newne < newce)
                    extend(newSets, newne, newce);
            }

            /* 
             * Upon return removal of the selected candidate from
             * compsub and its addition to the old set not (Step
             * 5)
             *
             * Remove from compsub */
            nel--;
            /* Add to not */
            ne++;
            if (DEBUG) System.out.println(prefix+"Incrementing ne, is now "+ne);

            /* 
             * Select a candidate disconnected to the fixed point (Step 1)
             */
            if( knod > 1) {  
                s = ne;
                while(edge[fixp][old[s]])
                    s++;
            }
            if (DEBUG) System.out.println(prefix+"next node is "+old[s]);
            if (s>ce) throw new RuntimeException("Clique finding failed. Do you have zeros in main diagonal?");
        } /* End of the loop with index knod */
	if (DEBUG) {
            System.out.println(prefix+"Returning from Recursion.");
            prefix = prefix.substring(0,prefix.length()-2);
	}
        return true;
    }


    static  TreeSet findHyperCliques(KHypergraph theGraph) {

        cliques = new TreeSet();

	int numNodes = theGraph.getNumNodes();
	int arity = theGraph.getArity();

        int nodes[] = new int[numNodes];
        for (int i=0; i<numNodes; i++)
            nodes[i]=i;
        
        findCliquesRec(theGraph, nodes,0);
        return cliques;
    }
    
    static private void findCliquesRec(KHypergraph theGraph, int nodes[], int recDepth) {
        if (recDepth>0 /*&& DEBUG*/ ) System.out.println("Recursion depth: "+recDepth);
	int numNodes = theGraph.getNumNodes();
	int arity = theGraph.getArity();
        
        Clique tmpClique;
	int cliqueCand[] = new int[numNodes];
	int clique[];
	int cliqueSize=0;
	KEdge tmpEdge, buildEdge,tmpEdge2;
	int edgeNodes[] = new int[arity];
	boolean goodNode,goodEdge;
	
	int tmpNodes[] = new int[nodes.length];
	int tmpNodeIndex=0;

        int iter;
	Vector edges;

	boolean atLeastOneGoodEdge;
        do {
            boolean terminate = false;

	    if (DEBUG)
		printArray("Subgraph: ",nodes);
	    
	    edges = theGraph.getAllEdgesOfSubgraph(nodes);
            boolean cliqueFound[] = new boolean[edges.size()];
	    atLeastOneGoodEdge=false;
	    // initialize with "false"
	    for (int i=0; i<cliqueFound.length; cliqueFound[i++]=false);
	    
            for (iter=0;iter<edges.size() && !terminate;iter++) {

                // take an edge
                tmpEdge = (KEdge)edges.elementAt(iter);
		if (DEBUG) System.out.println("Cliques for edge: "+tmpEdge);
                cliqueSize=0;
		
		goodEdge=true;
                for (int i=0; i<arity; i++) {
		    if (nodes[tmpEdge.nodeAt(i)]==-1) goodEdge=false;
                    cliqueCand[cliqueSize++]=tmpEdge.nodeAt(i); 
		}
		if (!goodEdge) continue;
                // find all nodes that are connected to tmpEdge in arity-1 nodes
                for (int i=0; i<numNodes; i++) { 
                    if (tmpEdge.contains(i)) continue; // not this one
                    if (!contains(nodes,i)) continue; // not this one either
                    // again relying on the correctness of that parameter
                    goodNode=true;
                    //now test if all necessary edges exist: replace
                    //each node in the edge with the new node and see
                    //if that edge exist. In fact, this will turn up
                    //nodes that are connected to each (arity-1)-set
                    //of nodes in tmpEdge by another edge
                    for (int j=0; j<arity; j++) {
                        System.arraycopy(tmpEdge.getEdge(),0,edgeNodes,0,arity);
                        edgeNodes[j]=i;
                        buildEdge = new KEdge(edgeNodes);
                        
                        if (!edges.contains(buildEdge)) {
                            goodNode=false;
                            break;
                        }
                    }
                    if (goodNode) {// node is good and should be included
                        cliqueCand[cliqueSize++]=i;
                    }
                }

                // now, clique contains a more or less connected set
                // of nodes that must be tested for the clique
                // property
                
                clique = new int[cliqueSize];
                System.arraycopy(cliqueCand,0,clique,0,cliqueSize);
                if (theGraph.isClique(clique)) {
		    // mark edge as "good", since clique has been found from it
		    cliqueFound[iter]=true;
		    atLeastOneGoodEdge = true;
		    
                    if (DEBUG) printArray("Clique: ",clique);

                    tmpClique = new Clique(clique);
		    boolean goodClique;
                    if (!cliques.contains(tmpClique)) {
			goodClique=true;
			for (Iterator iter3=cliques.iterator(); iter3.hasNext();) 
			    if (((Clique)iter3.next()).isSupersetOf(tmpClique)) {
				goodClique=false;
				break;
			    }
                        if (goodClique) {
			    cliques.add(tmpClique);
			    System.out.println("New clique (smart): "+tmpClique);
			}
                    }
		    else if (DEBUG) System.out.println("Clique already discovered: "+tmpClique);

                    // now test all nodes whether they have the
                    // mindegree necessary for the clique (in the
                    // complete graph!). Remove all nodes that do. If
                    // ALL nodes have the mindegree, we're done (we
                    // can return from this function), otherwise we'll
                    // terminate the loop by setting "terminate" to
                    // true and start again with the smaller
                    // graph. Let's see what that does.
                    
		    /*
		    int mindegree = binomial(cliqueSize-1,arity-1);
                    int alldegrees[] = theGraph.getAllDegreesOfSubgraph(nodes);
                    
                    if (DEBUG) System.out.println("mindegree is "+mindegree);
                    for (int i=0; i<nodes.length; i++)
                        if (alldegrees[i]!=mindegree  // higher or lower.
                            ||
                            !contains(clique,nodes[i])) {// or not in clique
                            // alldegrees[i] is the degree of the node
                            // at position i in nodes[]

			    // this was creating a new array with only the valid nodes.
                            //tmpNodes[tmpNodeIndex++]=nodes[i];
			    
			    // let's try somethign different and mark
			    // deleted nodes with -1
			    if (nodes[i]>=0) tmpNodeIndex++;// we still need to count for the break-condition
			}
			else { 
			    if (DEBUG) System.out.println("Removing: "+nodes[i]);
			    nodes[i]=-1;
			}
		    */
		    
		    /*
		    edges.removeElement(tmpEdge);
		    System.out.println("Removing Edge: "+tmpEdge); 
		    iter--;
		    */

			
			     /*     
                    if (tmpNodeIndex==0){
			if (isRecursion && DEBUG) System.out.println("Return from recursion");
			return;
		    }
			     */
		    /*
                    if (tmpNodeIndex<nodes.length) {
                        nodes = new int[tmpNodeIndex];
                        for (int i=0; i<tmpNodeIndex; i++)
                            nodes[i]=tmpNodes[i];
                        terminate=true;
                    }
                    */
                    
                }
                else {
                    if (DEBUG) printArray("No Clique: ",clique);
		    /*
		    for (int i=0; i<nodes.length; i++)
			tmpNodes[i]=-1;
		    for(int i=0; i<clique.length; i++)
			tmpNodes[clique[i]]=clique[i];
                    if (!isRecursion) findCliquesRec(theGraph,tmpNodes,true);
		    */
                }
                //need to remove all nodes that have minimum clique
                //degree from the graph
            }
	    
	    if (!atLeastOneGoodEdge) {
		// we have no good edge. All edges are parts of cliques
		// we need to split the graph

		tmpEdge = (KEdge)edges.elementAt(0); // a random edge

		int firstNodes[] = new int[nodes.length];
		int secondNodes[] = new int[nodes.length];
		int numFirstNodes =0;

		// initialize
		for (int i=0; i<nodes.length; i++) {
		    firstNodes[i]=-1;
		    secondNodes[i]=i;
		}
		
		tmpNodes = tmpEdge.getEdge();
		for (int i=0; i<tmpEdge.getArity(); i++) {
		    firstNodes[tmpNodes[i]]=tmpNodes[i];
		    numFirstNodes++;
		    secondNodes[tmpNodes[i]]=-1;
		}
                // find all nodes that are connected to tmpEdge in arity-1 nodes
		boolean atLeastOneBadNode = false;
                for (int i=0; i<numNodes; i++) { 
                    if (tmpEdge.contains(i)) continue; // not this one
                    if (!contains(nodes,i)) continue; // not this one either
                    // again relying on the correctness of that parameter
                    goodNode=true;
                    //now test if all necessary edges exist: replace
                    //each node in the edge with the new node and see
                    //if that edge exist. In fact, this will turn up
                    //nodes that are connected to each (arity-1)-set
                    //of nodes in tmpEdge by another edge
                    for (int j=0; j<arity; j++) {
                        System.arraycopy(tmpEdge.getEdge(),0,edgeNodes,0,arity);
                        edgeNodes[j]=i;
                        buildEdge = new KEdge(edgeNodes);
                        
                        if (!edges.contains(buildEdge)) {
                            goodNode=false;
                            break;
                        }
                    }
                    if (goodNode) {
			// node is good and should be included in
			// first graph but removed from second graph
			firstNodes[i]=i;
			numFirstNodes++;
			secondNodes[i]=-1;
                    }
		    else atLeastOneBadNode=true;
                }

		// now we have to "pack" those two arrays to make
		// edges for subedge testing.

		int packedFirstNodes[] = new int[numFirstNodes];
		int packedSecondNodes[] = new int[numNodes-numFirstNodes];
		
		int firstIndex=0;
		int secondIndex=0;
		for(int i=0; i<numNodes; i++) {
		    if (firstNodes[i]!=-1)
			packedFirstNodes[firstIndex++]=firstNodes[i];
		    if (secondNodes[i]!=-1)
			packedSecondNodes[secondIndex++]=secondNodes[i];
		}
		
		KEdge firstEdge = new KEdge(packedFirstNodes);
		// hmm, the second graph must still have all its nodes...
		KEdge secondEdge = new KEdge(packedSecondNodes);
		
		if (DEBUG) System.out.println("Temp.Edge:"+ firstEdge);//+ ", "+secondEdge);

		KHypergraph firstGraph = new KHypergraph(arity,numNodes);
		KHypergraph secondGraph = new KHypergraph(arity,numNodes);
		// first graph contains all edges that have arity-1
		// nodes in common with tmpEdge
		edgeNodes = tmpEdge.getEdge();
		//tmpNodes = new int[tmpEdge.getArity()-1];

		// now add only edges belonging to tmpEdge to firstGraph.
		// secondGraph has ALL edges except tmpEdge
		for (Iterator iter2=edges.iterator(); iter2.hasNext();) {
		    tmpEdge2 = (KEdge)iter2.next();
		    if (tmpEdge2.isSubsetOf(firstEdge))
			    firstGraph.addEdge(tmpEdge2.getEdge());
		    if (!tmpEdge2.equals(tmpEdge))
			secondGraph.addEdge(tmpEdge2.getEdge());
		    
		}
		    
		//System.arraycopy(nodes,0,tmpNodes,0,nodes.length);


	        if (!atLeastOneBadNode) {// graph can't be reduced
		    TreeSet tmpCliques = findHyperCliquesBruteForce(firstGraph);
		    boolean contained = false;
		    Clique tmpClique2;
		    for (Iterator iter2 = tmpCliques.iterator(); iter2.hasNext();) {
			tmpClique = (Clique)iter2.next();
			contained=false;
			for (Iterator iter3 = cliques.iterator(); iter3.hasNext();) {
			    tmpClique2 = (Clique)iter3.next();
			    if (tmpClique2.isSupersetOf(tmpClique)) {
				contained=true;
				break;
			    }
			}
			if (!contained)
			    cliques.add(tmpClique);
		    }
		}
		else {
		    if (DEBUG) System.out.println("exploring first graph ("+firstGraph.edges.size()+" edges)");
		    findCliquesRec(firstGraph,firstNodes,recDepth+1);
		    if (DEBUG) System.out.println("exploring second graph ("+secondGraph.edges.size()+" edges)");
		    
		    findCliquesRec(secondGraph,nodes,recDepth+1);
		}
	    }

	    // now retain only good edges in the graph
	    KHypergraph tmpGraph = new KHypergraph(arity,numNodes);
	    if (DEBUG) {
		System.out.print("Cliques found: ");
		for (int i=0; i<cliqueFound.length; i++)
		    System.out.print(cliqueFound[i]+" ");
		System.out.println();
	    }

	    for (int i =0; i<edges.size(); i++) {
		if (!cliqueFound[i])
		    tmpGraph.addEdge(((KEdge)edges.elementAt(i)).getEdge());
	    }
	    // and use the new graph
	    theGraph = tmpGraph;
	    
        }
        while(theGraph.edges.size()>0 && atLeastOneGoodEdge);
	if (recDepth>0 && DEBUG) System.out.println("Return from recursion");
    }

    /** this function finds all hypercliques in a k-Hypergraph. It
        is terribly inefficient but the best I could do. */
    static TreeSet findHyperCliquesBruteForce(KHypergraph theGraph) {
        TreeSet cliques=new TreeSet();
        int highest=-1;
        int arity=theGraph.getArity();

        int numberOfNodes = theGraph.getNumNodes(); 


        int degrees[] = theGraph.getAllDegrees();

        if (DEBUG) {
            System.out.println("Node degrees are:");
            for (int i =0; i<numberOfNodes;i++) 
                // numberOfNodes should be equal to degrees.length
                System.out.print(degrees[i]+" ");
            System.out.println();
        }

        // find highest node degree
        int degree;
        for (int node=0; node<numberOfNodes; node++) {
            degree=degrees[node];
            if (degree>highest) highest=degree;
        }
        if (DEBUG) System.out.println("Highest node degree: "+highest);
    

	// the maximum clique size is bounded somehow by the maximum
	// node degree. In fact, for a k-Hyperclique of n nodes one
	// needs (n choose k) k-hyperedges. Each of them connects k
	// nodes, so the degree of each node must be at least (n
	// choose k)*k/n = (n-1 choose k-1). 

	int maxCliqueSize = antibinomial(arity-1,highest)-1; 
        
        if (DEBUG) System.out.println("Upper bound for clique size is: "+maxCliqueSize);
	
        // but we need (n-1 choose k-1) nodes of degree n to form a
        // clique on a k-Hypergraph
        boolean enoughNodesWithMinDegree=false;
        int minDegree,  numNodesWithMinDegree;
        do {
            minDegree = binomial(maxCliqueSize-1,arity-1);
	    if (DEBUG) System.out.println("The minDegree is: "+minDegree);
            // check all nodes for this degree
            numNodesWithMinDegree=0;
            for(int i =0; i<numberOfNodes; i++)
                if(degrees[i]>=minDegree) numNodesWithMinDegree++;
            if(numNodesWithMinDegree>=maxCliqueSize) enoughNodesWithMinDegree=true;
            else maxCliqueSize--;
        }
        while(!enoughNodesWithMinDegree && maxCliqueSize>=arity);

	// now maxCliqueSize is the maximal clique size that could be
	// explained with the node degrees in the graph
            
	if (DEBUG) {
            if (maxCliqueSize>=arity)
                System.out.println("Maximum cliqueSize is "+maxCliqueSize);
            else
                System.out.println("No large enough cliques found.");
        }

        // generate all subsets of sizes between max and min possible
        int cliqueSize;
	// a new set, goodIndices holds only those nodes that have a
	// sufficient degree for the current value of cliqueSize
        int goodIndices[]=new int[numberOfNodes];
        int numOfGoodIndices;
        for (cliqueSize=maxCliqueSize; cliqueSize>=arity; cliqueSize--) {
            // cliques with cliqueSize>arity, less doesn't make sense

	    if (DEBUG) System.out.println("Checking cliques of size: "+cliqueSize);
            // now, create a new array "indices" which has only the
            // nodes with the appropriate minimum degree and use that
            // for the subset generation.

            minDegree = binomial(cliqueSize-1,arity-1);
	    
            //grrrrrr.....that's not trivial. We DO have to find a
            //good hyperclique finding algorithm...
	    numOfGoodIndices=0;
            for (int i=0; i<numberOfNodes; i++)
		if (degrees[i]>=minDegree) 
		    goodIndices[numOfGoodIndices++]=i; 
            // note that this can become larger than cliquesize!

	    if (DEBUG) {
		System.out.print("Good Indices: ");
		for (int i=0; i<numOfGoodIndices; i++)
		    System.out.print(goodIndices[i]+" ");
		System.out.println();
	    }
	    
            // form all cliqueSize-subsets of nodes (but only those
            // whose node numbers are in lexicographic order)
            int indices[] = new int[cliqueSize]; 
            // for the indices of the subset elements

	    // we also need another level of indirection, for the
	    // translation from indices to goodIndices
	    int translatedIndices[] = new int[cliqueSize];
            
            // initialize indices
            for (int i = 0; i<cliqueSize; i++)
                indices[i]=i; 

            
            int i_ptr = cliqueSize-1;
            // now form subsets 

            boolean cliqueFound,cliqueIsSubset;
            Clique tmpClique;
            do {

                // avoid testing subsets of existing cliques
                cliqueIsSubset=false;
		for(int i=0; i<cliqueSize; i++)
		    translatedIndices[i]=goodIndices[indices[i]];

                tmpClique = new Clique(translatedIndices);
                for (Iterator iter=cliques.iterator(); iter.hasNext();) {
                    if(((Clique)iter.next()).isSupersetOf(tmpClique)) {
                        cliqueIsSubset=true;
                        break;
                    }
                }

                // test for cliqueness
                if (!cliqueIsSubset)
                    cliqueFound = theGraph.isClique(translatedIndices);
                else cliqueFound=false;

                /* // too much output
                if (DEBUG) { 
                    if (DEBUG) System.out.print("Subset: ");
                    for (int m=0; m<cliqueSize; m++)
                        System.out.print(indices[m]+" ");
                    System.out.println();
                }
                */
                
                boolean contained;
                Clique tmpClique2;
                if (cliqueFound) {
                    tmpClique=new Clique(translatedIndices);
                    contained=false;
                    for (Iterator iter3 = cliques.iterator(); iter3.hasNext();) {
                        tmpClique2 = (Clique)iter3.next();
                        if (tmpClique2.isSupersetOf(tmpClique)) {
                            contained=true;
                            break;
                        }
                    }
                    if (!contained) {
                        System.out.println("New clique (brute force): "+tmpClique);
                        cliques.add(tmpClique);
                    }
                }
                
                
                // generate next subset
                i_ptr=cliqueSize-1;
                indices[i_ptr]++;

                // if out of bounds, push up the carry-overs
                for (i_ptr=cliqueSize-1;i_ptr>0;i_ptr--) {
                    if (indices[i_ptr]==numOfGoodIndices-(cliqueSize-1-i_ptr)) {
			// index at position k cannot be larger than
			// k+(numOfGoodIndices (was:
			// numberOfNodes)-cliqueSize)
                        if (i_ptr==0) break;
                        else {
                            indices[i_ptr-1]++;
                            indices[i_ptr]=-1;
                        }
                    }
                }
                for(i_ptr=1; i_ptr<cliqueSize; i_ptr++) {
                    if (indices[i_ptr]==-1)
                        indices[i_ptr]=indices[i_ptr-1]+1;
                }
            }
            while (indices[0]<numOfGoodIndices-(cliqueSize-1));

        }
        return cliques;
    }


    /** array containment test */
    static private boolean contains(int [] array, int element) {
        for (int i=0; i<array.length; i++)
            if (array[i]==element) return true;
        return false;
    }

    static void printArray(String leadIn, int array[]) {
	System.out.print(leadIn);
	for (int i=0; i<array.length; i++)
	    System.out.print(array[i]+" ");
	System.out.println();
    }


    /** binomial coefficient of n and k */
    static private int binomial(int n, int k) {
        if (n<k) return -1; // error

        int result =1;
        for(int i = 0; i< k; i++)
            result *= n-i;
        result /= fact(k);
        return result;
    }

    
    /** returns the largest n for which binom >= (n choose k) */
    public static int antibinomial(int k, int binom) {
        if (binom<1) return 0;

        // k is constant
        long result=1;
        int tmpn=k+1;
        long kFact = fact(k);
        do {
            result*=tmpn;
            result/=tmpn-k;
            tmpn++;
        }
        while(binom>=result);
        return tmpn--;
 
    }
    
    /** factorial function */
    public static int fact(int x) {
        int result=1;
        while(x>1) result*=x--;
        return result;
    }
}
/******************************************************************/

